from collections.abc import Mapping
from dataclasses import MISSING
from dataclasses import Field as _Field
from typing import NewType

from ._type_def import ExplicitNull
from .constants import PY310_OR_ABOVE, PY314_OR_ABOVE
from .utils._object_path import split_object_path

# Define a simple type (alias) for the `CatchAll` field
#
# The `type` statement is introduced in Python 3.12
# Ref: https://docs.python.org/3.12/reference/simple_stmts.html#type
#
# TODO: uncomment following usage of `type` statement
#   once we drop support for Python 3.9 - 3.11
# if PY312_OR_ABOVE:
#     type CatchAll = Mapping
CatchAll = NewType('CatchAll', Mapping)


def _normalize_alias_path_args(all_paths, load, dump):
    """Normalize `AliasPath` arguments and canonicalize path values."""
    if load is not None:
        all_paths = load
        load = None
        dump = ExplicitNull

    elif dump is not None:
        all_paths = dump
        dump = None
        load = ExplicitNull

    if isinstance(all_paths, str):
        all_paths = (split_object_path(all_paths),)
    else:
        all_paths = tuple([
            split_object_path(a) if isinstance(a, str) else a
            for a in all_paths
        ])

    return all_paths, load, dump


def _normalize_alias_args(default, default_factory, all_aliases, load, dump, env):
    """Normalize `Alias` arguments and canonicalize alias values."""

    if default is not MISSING and default_factory is not MISSING:
        raise ValueError('cannot specify both default and default_factory')

    if all_aliases:
        load = dump = all_aliases

    elif load is not None and isinstance(load, str):
        load = (load,)

    elif env is not None:
        if isinstance(env, str):
            env = (env,)
        elif env is True:
            env = load

    return all_aliases, load, dump, env


# Instances of Field are only ever created from within this module,
# and only from the field() function, although Field instances are
# exposed externally as (conceptually) read-only objects.
#
# name and type are filled in after the fact, not in __init__.
# They're not known at the time this class is instantiated, but it's
# convenient if they're available later.

# noinspection PyPep8Naming,PyShadowingBuiltins
def Env(*load,
        default=MISSING,
        default_factory=MISSING,
        init=True, repr=True,
        hash=None, compare=True, metadata=None,
        **field_kwargs):

    # noinspection PyTypeChecker
    return Alias(
        env=load,
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
        **field_kwargs,
    )

# In Python 3.14, dataclasses adds a new parameter to the :class:`Field`
# constructor: `doc`
#
# Ref: https://docs.python.org/3.14/library/dataclasses.html#dataclasses.field
if PY314_OR_ABOVE:
    # noinspection PyPep8Naming,PyShadowingBuiltins
    def Alias(
        *all,
        load=None,
        dump=None,
        env=None,
        skip=False,
        default=MISSING,
        default_factory=MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None,
        kw_only=MISSING,
        doc=None,
    ):

        all, load, dump, env = _normalize_alias_args(default, default_factory, all, load, dump, env)

        return Field(
            load,
            dump,
            env,
            skip,
            None,
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
            kw_only,
            doc,
        )

    # noinspection PyPep8Naming,PyShadowingBuiltins
    def AliasPath(
        *all,
        load=None,
        dump=None,
        skip=False,
        default=MISSING,
        default_factory=MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None,
        kw_only=MISSING,
        doc=None,
    ):
        all, load, dump = _normalize_alias_path_args(all, load, dump)

        return Field(
            load,
            dump,
            load,
            skip,
            all,
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
            kw_only,
            doc,
        )

    # noinspection PyShadowingBuiltins
    def skip_if_field(
        condition,
        *,
        default=MISSING,
        default_factory=MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None,
        kw_only=MISSING,
        doc=None,
    ):

        if default is not MISSING and default_factory is not MISSING:
            raise ValueError("cannot specify both default and default_factory")

        if metadata is None:
            metadata = {}

        metadata["__skip_if__"] = condition

        return _Field(
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
            kw_only,
            doc,
        )

    class Field(_Field):

        __slots__ = ("load_alias", "dump_alias", "env_vars", "skip", "path")

        # noinspection PyShadowingBuiltins
        def __init__(
            self,
            load_alias,
            dump_alias,
            env_vars,
            skip,
            path,
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
            kw_only,
            doc=None,
        ):

            # noinspection PyArgumentList
            super().__init__(
                default,
                default_factory,
                init,
                repr,
                hash,
                compare,
                metadata,
                kw_only,
                doc,
            )

            self.load_alias = load_alias
            self.dump_alias = dump_alias
            self.env_vars = env_vars
            self.skip = skip
            self.path = path


# In Python 3.10, dataclasses adds a new parameter to the :class:`Field`
# constructor: `kw_only`
#
# Ref: https://docs.python.org/3.10/library/dataclasses.html#dataclasses.dataclass
elif PY310_OR_ABOVE:  # pragma: no cover

    # noinspection PyPep8Naming,PyShadowingBuiltins
    def Alias(*all,
              load=None,
              dump=None,
              env=None,
              skip=False,
              default=MISSING,
              default_factory=MISSING,
              init=True, repr=True,
              hash=None, compare=True,
              metadata=None, kw_only=MISSING):

        all, load, dump, env = _normalize_alias_args(default, default_factory, all, load, dump, env)

        return Field(
            load,
            dump,
            env,
            skip,
            None,
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
            kw_only,
        )

    # noinspection PyPep8Naming,PyShadowingBuiltins
    def AliasPath(*all,
                  load=None,
                  dump=None,
                  skip=False,
                  default=MISSING,
                  default_factory=MISSING,
                  init=True, repr=True,
                  hash=None, compare=True,
                  metadata=None, kw_only=MISSING):
        all, load, dump = _normalize_alias_path_args(all, load, dump)

        return Field(
            load,
            dump,
            load,
            skip,
            all,
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
            kw_only,
        )

    # noinspection PyShadowingBuiltins
    def skip_if_field(
        condition,
        *,
        default=MISSING,
        default_factory=MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None,
        kw_only=MISSING
    ):

        if default is not MISSING and default_factory is not MISSING:
            raise ValueError("cannot specify both default and default_factory")

        if metadata is None:
            metadata = {}

        metadata["__skip_if__"] = condition

        return _Field(
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
            kw_only,
        )

    class Field(_Field):

        __slots__ = ('load_alias',
                     'dump_alias',
                     'env_vars',
                     'skip',
                     'path')

        # noinspection PyShadowingBuiltins
        def __init__(self,
                     load_alias, dump_alias, env_vars, skip, path,
                     default, default_factory, init, repr, hash, compare,
                     metadata, kw_only):

            super().__init__(default, default_factory, init, repr, hash,
                             compare, metadata, kw_only)

            if path is not None:
                if isinstance(path, str):
                    path = split_object_path(path) if path else (path, )

            self.load_alias = load_alias
            self.dump_alias = dump_alias
            self.env_vars = env_vars
            self.skip = skip
            self.path = path

else:  # pragma: no cover
    # noinspection PyPep8Naming,PyShadowingBuiltins
    def Alias(*all,
              load=None,
              dump=None,
              env=None,
              skip=False,
              default=MISSING,
              default_factory=MISSING,
              init=True, repr=True,
              hash=None, compare=True, metadata=None):

        all, load, dump, env = _normalize_alias_args(default, default_factory, all, load, dump, env)

        return Field(
            load,
            dump,
            env,
            skip,
            None,
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
        )

    # noinspection PyPep8Naming,PyShadowingBuiltins
    def AliasPath(*all,
                  load=None,
                  dump=None,
                  skip=False,
                  default=MISSING,
                  default_factory=MISSING,
                  init=True, repr=True,
                  hash=None, compare=True,
                  metadata=None):
        all, load, dump = _normalize_alias_path_args(all, load, dump)

        return Field(
            load,
            dump,
            load,
            skip,
            all,
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
        )

    # noinspection PyShadowingBuiltins
    def skip_if_field(
        condition,
        *,
        default=MISSING,
        default_factory=MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None
    ):

        if default is not MISSING and default_factory is not MISSING:
            raise ValueError("cannot specify both default and default_factory")

        if metadata is None:
            metadata = {}

        metadata["__skip_if__"] = condition

        # noinspection PyArgumentList
        return _Field(
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
        )

    class Field(_Field):

        __slots__ = ('load_alias',
                     'dump_alias',
                     'env_vars',
                     'skip',
                     'path')

        # noinspection PyArgumentList,PyShadowingBuiltins
        def __init__(self,
                     load_alias, dump_alias, env_vars, skip, path,
                     default, default_factory, init, repr, hash, compare,
                     metadata):

            super().__init__(default, default_factory, init, repr, hash,
                             compare, metadata)

            if path is not None:
                if isinstance(path, str):
                    path = split_object_path(path) if path else (path,)

            self.load_alias = load_alias
            self.dump_alias = dump_alias
            self.env_vars = env_vars
            self.skip = skip
            self.path = path


Alias.__doc__ = """
    Maps one or more JSON key names to a dataclass field.

    This function acts as an alias for ``dataclasses.field(...)``, with additional
    support for associating a field with one or more JSON keys. It customizes
    serialization and deserialization behavior, including handling keys with
    varying cases or alternative names.

    The mapping is case-sensitive; JSON keys must match exactly (e.g., ``myField``
    will not match ``myfield``). If multiple keys are provided, the first one
    is used as the default for serialization.

    :param all: One or more JSON key names to associate with the dataclass field.
    :type all: str
    :param load: Key(s) to use for deserialization. Defaults to ``all`` if not specified.
    :type load: str | Sequence[str] | None
    :param dump: Key to use for serialization. Defaults to the first key in ``all``.
    :type dump: str | None
    :param skip: If ``True``, the field is excluded during serialization. Defaults to ``False``.
    :type skip: bool
    :param default: Default value for the field. Cannot be used with ``default_factory``.
    :type default: Any
    :param default_factory: Callable to generate the default value. Cannot be used with ``default``.
    :type default_factory: Callable[[], Any]
    :param init: Whether the field is included in the generated ``__init__`` method. Defaults to ``True``.
    :type init: bool
    :param repr: Whether the field appears in the ``__repr__`` output. Defaults to ``True``.
    :type repr: bool
    :param hash: Whether the field is included in the ``__hash__`` method. Defaults to ``None``.
    :type hash: bool
    :param compare: Whether the field is included in comparison methods. Defaults to ``True``.
    :type compare: bool
    :param metadata: Additional metadata for the field. Defaults to ``None``.
    :type metadata: dict
    :param kw_only: If ``True``, the field is keyword-only. Defaults to ``False``.
    :type kw_only: bool
    :return: A dataclass field with additional mappings to one or more JSON keys.
    :rtype: Field

    **Examples**

    **Example 1** -- Mapping multiple key names to a field::

        from dataclasses import dataclass

        from dataclass_wizard import Alias, LoadMeta, fromdict

        @dataclass
        class Example:
            my_field: str = Alias('key1', 'key2', default="default_value")

        print(fromdict(Example, {'key2': 'a value!'}))
        #> Example(my_field='a value!')

    **Example 2** -- Skipping a field during serialization::

        from dataclasses import dataclass

        from dataclass_wizard import Alias, JSONWizard

        @dataclass
        class Example(JSONWizard):
            my_field: str = Alias('key', skip=True)

        ex = Example.from_dict({'key': 'some value'})
        print(ex)                  #> Example(my_field='a value!')
        assert ex.to_dict() == {}  #> True
"""

AliasPath.__doc__ = """
    Creates a dataclass field mapped to one or more nested JSON paths.

    This function acts as an alias for ``dataclasses.field(...)``, with additional
    functionality to associate a field with one or more nested JSON paths,
    including complex or deeply nested structures.

    The mapping is case-sensitive, meaning that JSON keys must match exactly
    (e.g., "myField" will not match "myfield"). Nested paths can include dot
    notations or bracketed syntax for accessing specific indices or keys.

    :param all: One or more nested JSON paths to associate with
        the dataclass field (e.g., ``a.b.c`` or ``a["nested"]["key"]``).
    :type all: PathType | str
    :param load: Path(s) to use for deserialization. Defaults to ``all`` if not specified.
    :type load: PathType | str | None
    :param dump: Path(s) to use for serialization. Defaults to ``all`` if not specified.
    :type dump: PathType | str | None
    :param skip: If True, the field is excluded during serialization. Defaults to False.
    :type skip: bool
    :param default: Default value for the field. Cannot be used with ``default_factory``.
    :type default: Any
    :param default_factory: A callable to generate the default value. Cannot be used with ``default``.
    :type default_factory: Callable[[], Any]
    :param init: Whether the field is included in the generated ``__init__`` method. Defaults to True.
    :type init: bool
    :param repr: Whether the field appears in the ``__repr__`` output. Defaults to True.
    :type repr: bool
    :param hash: Whether the field is included in the ``__hash__`` method. Defaults to None.
    :type hash: bool
    :param compare: Whether the field is included in comparison methods. Defaults to True.
    :type compare: bool
    :param metadata: Additional metadata for the field. Defaults to None.
    :type metadata: dict
    :param kw_only: If True, the field is keyword-only. Defaults to False.
    :type kw_only: bool
    :return: A dataclass field with additional mapping to one or more nested JSON paths.
    :rtype: Field

    **Examples**

    **Example 1** -- Mapping multiple nested paths to a field::

        from dataclasses import dataclass

        from dataclass_wizard import AliasPath, fromdict

        @dataclass
        class Example:
            my_str: str = AliasPath('a.b.c.1', 'x.y["-1"].z', default="default_value")

        # Maps nested paths ('a', 'b', 'c', 1) and ('x', 'y', '-1', 'z')
        # to the `my_str` attribute. '-1' is treated as a literal string key,
        # not an index, for the second path.

        print(fromdict(Example, {'x': {'y': {'-1': {'z': 'some_value'}}}}))
        #> Example(my_str='some_value')

    **Example 2** -- Using Annotated::

        from dataclasses import dataclass
        from typing import Annotated

        from dataclass_wizard import AliasPath, JSONWizard

        @dataclass
        class Example(JSONWizard):
            my_str: Annotated[str, AliasPath('my."7".nested.path.-321')]


        ex = Example.from_dict({'my': {'7': {'nested': {'path': {-321: 'Test'}}}}})
        print(ex)  #> Example(my_str='Test')
"""

Field.__doc__ = """
    Alias to a :class:`dataclasses.Field`, but one which also represents a
    mapping of one or more JSON key names to a dataclass field.

    See the docs on the :func:`Alias` and :func:`AliasPath` for more info.
"""



from collections.abc import Sequence
from os import PathLike

from ._env import E

SecretsDir = str | PathLike[str]
SecretsDirs = SecretsDir | Sequence[SecretsDir] | None

Environ = dict[str, 'str | None']
SecretsFileMapping = dict[str, str]

EnvFilePath = str | PathLike[str]
EnvFilePaths = bool | EnvFilePath | Sequence[EnvFilePath] | None

def get_secrets_map(cls: E, secret_dirs: SecretsDirs, *, reload: bool = False) -> SecretsFileMapping: ...
def get_dotenv_map(cls: E, env_file: EnvFilePaths, *, reload: bool = False) -> Environ: ...

def read_secrets_dirs(dirs: SecretsDirs) -> SecretsFileMapping: ...
def dotenv_values(files: EnvFilePaths) -> Environ: ...



from __future__ import annotations

import collections.abc as abc
from base64 import b64encode
from collections import defaultdict, deque
from collections.abc import Collection
from dataclasses import MISSING, Field, is_dataclass
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from pathlib import Path

# noinspection PyUnresolvedReferences,PyProtectedMember
from typing import (
    Any,
    Callable,
    Literal,
    NamedTuple,
    cast,
)
from uuid import UUID

from ._bases import AbstractMeta, BaseDumpHook
from ._class_helper import (
    CLASS_TO_DUMPER,
    DATACLASS_FIELD_TO_ALIAS_PATH_FOR_DUMP,
    dataclass_field_to_skip_if,
    resolve_dataclass_field_to_alias_for_dump,
    set_class_dumper,
)
from ._decorators import (
    setup_recursive_safe_function,
    setup_recursive_safe_function_for_generic,
)
from ._log import LOG
from ._meta_cache import get_meta
from ._models import (
    LEAF_TYPES,
    LEAF_TYPES_NO_BYTES,
    Extras,
    TypeInfo,
    finalize_skip_if,
    get_skip_if_condition,
)
from ._models_date import UTC, ZERO
from ._type_conv import datetime_to_timestamp
from ._type_def import (
    META,
    ExplicitNull,
    JSONObject,
    NoneType,
    PyLiteralString,
    T,
)
from ._type_utils import create_new_class, is_subclass_safe

# noinspection PyUnresolvedReferences
from .constants import _HOOKS, CATCH_ALL, PACKAGE_NAME, TAG
from .enums import DateTimeTo, KeyCase
from .errors import JSONWizardError, MissingData, MissingFields, ParseError
from .utils._dataclass_compat import (
    SEEN_DEFAULT,
    dataclass_field_names,
    dataclass_fields,
    set_new_attribute,
)
from .utils._dict_helper import NestedDict
from .utils._function_builder import FunctionBuilder
from .utils._typing_compat import (
    eval_forward_ref_if_needed,
    get_args,
    get_keys_for_typed_dict,
    get_origin_v2,
    is_annotated,
    is_typed_dict,
    is_typed_dict_type_qualifier,
    is_union,
)

_KNOWN_FACTORY_LITERALS: dict[Callable[[], Any], str] = {
    list: '[]',
    dict: '{}',
    set: 'set()',
    tuple: '()',
    frozenset: 'frozenset()',
}


def default_compare_expr(
    f: Field[Any],
    locals_ns: dict[str, Any],
    default_name: str,
    *,
    allow_calling_unknown_factories: bool = True,
) -> str | None:
    """
    Return an expression string to compare against for default-elision.
    None means: cannot/should not elide.
    """
    # scalar/object default: bind it and reference by name
    if f.default is not MISSING:
        locals_ns[default_name] = f.default
        return default_name

    df = f.default_factory
    if df is not MISSING:
        lit = _KNOWN_FACTORY_LITERALS.get(df)
        if lit is not None:
            return lit

        if allow_calling_unknown_factories:
            locals_ns[default_name] = df
            return f'{default_name}()'

        # safest default (in case of non-deterministic factories)
        return None

    return None


def _type_returns_value_unchanged(arg, leaf_handling_as_subclass, origin=None):
    # scalar type:
    # (str, int, float, bool, complex, type, Literal, Any)
    if origin is None:
        origin = get_origin_v2(arg)
    return (origin is Any
            or origin is Literal
            or origin in LEAF_TYPES_NO_BYTES
            or (leaf_handling_as_subclass
                and is_subclass_safe(origin, LEAF_TYPES_NO_BYTES)))


def _all_return_value_unchanged(args, leaf_handling_as_subclass):
    for arg in args:
        if not _type_returns_value_unchanged(arg, leaf_handling_as_subclass):
            return False
    return True


class DumpMixin(BaseDumpHook):
    """
    This Mixin class derives its name from the eponymous `json.dumps`
    function. Essentially it contains helper methods to convert a `dataclass`
    to JSON strings (or a Python dictionary object).

    Refer to the :class:`AbstractDumper` class for documentation on any of the
    implemented methods.

    """
    __slots__ = ()

    def __init_subclass__(cls, _setup_defaults=True, **kwargs):
        super().__init_subclass__(**kwargs)
        if _setup_defaults:
            setup_default_dumper(cls)

    transform_dataclass_field = None

    @staticmethod
    def dump_fallback(tp: TypeInfo, _extras: Extras):
        # identity: o
        return tp.v()

    dump_from_str = dump_fallback
    dump_from_int = dump_fallback
    dump_from_float = dump_fallback
    dump_from_bool = dump_fallback
    dump_from_literal = dump_fallback

    @staticmethod
    def dump_from_bytes(tp: TypeInfo, extras: Extras):
        tp.ensure_in_locals(extras, b64encode)
        return f"b64encode({tp.v()}).decode('ascii')"

    @classmethod
    def dump_from_bytearray(cls, tp: TypeInfo, extras: Extras):
        tp.ensure_in_locals(extras, b64encode)
        return f"b64encode(bytes({tp.v()})).decode('ascii')"

    @staticmethod
    def dump_from_none(tp: TypeInfo, extras: Extras):
        return 'None'

    @staticmethod
    def dump_from_enum(tp: TypeInfo, extras: Extras):
        # alias: o.value
        return f'{tp.v()}.value'

    @staticmethod
    def dump_from_uuid(tp: TypeInfo, extras: Extras):
        # alias: o.hex
        return f'{tp.v()}.hex'

    @classmethod
    def dump_from_iterable(cls, tp: TypeInfo, extras: Extras):
        v, v_next, i_next = tp.v_and_next()
        gorg = tp.origin

        if v_next[0] == 'k':
            # raise same error as `json` (not serializable)
            raise TypeError('keys must be str, int, float, bool or None, '
                            f'not {gorg.__qualname__}') from None

        # noinspection PyBroadException
        try:
            elem_type = tp.args[0]
        except:
            elem_type = Any

        string = cls.dump_dispatcher_for_annotation(
            tp.replace(origin=elem_type, i=i_next, index=None, val_name=None), extras)

        if string == v_next:
            return f'{v}.copy()' if issubclass(gorg, list) else f'list({v})'

        return f'[{string} for {v_next} in {v}]'

    @classmethod
    def dump_from_tuple(cls, tp: TypeInfo, extras: Extras):
        args = tp.args

        # Determine the code string for the annotation

        # Check if the `Tuple` appears in the variadic form
        #   i.e. Tuple[str, ...]
        if args:
            is_variadic = args[-1] is ...
        else:
            # Annotated without args, as simply `tuple`
            return f'list({tp.v()})'

        if is_variadic:
            # Logic that handles the variadic form of :class:`Tuple`'s,
            # i.e. ``Tuple[str, ...]``
            #
            # Per `PEP 484`_, only **one** required type is allowed before the
            # ``Ellipsis``. That is, ``Tuple[int, ...]`` is valid whereas
            # ``Tuple[int, str, ...]`` would be invalid. `See here`_ for more info.
            #
            # .. _PEP 484: https://www.python.org/dev/peps/pep-0484/
            # .. _See here: https://github.com/python/typing/issues/180
            v, v_next, i_next = tp.v_and_next()

            # Given `Tuple[T, ...]`, we only need the generated string for `T`
            string = cls.dump_dispatcher_for_annotation(
                tp.replace(origin=args[0], i=i_next, index=None, val_name=None), extras)

            if string == v_next:
                return f'list({v})'

            result = f'[{string} for {v_next} in {v}]'

        else:
            string = ', '.join([
                str(cls.dump_dispatcher_for_annotation(
                    tp.replace(origin=arg, index=k),
                    extras))
                for k, arg in enumerate(args)])

            result = f'[{string}]'

        return result

    @classmethod
    def dump_from_named_tuple(cls, tp: TypeInfo, extras: Extras):
        nt_tp = cast(NamedTuple, tp.origin)
        fields = nt_tp._fields  # field names in order
        ann = nt_tp.__annotations__

        field_to_value = {
            name: str(
                cls.dump_dispatcher_for_annotation(
                    tp.replace(origin=ann.get(name, Any), index=i),
                    extras,
                )
            )
            for i, name in enumerate(fields)
        }

        if extras['config'].namedtuple_as_dict:
            params = [f'{field!r}: {value}' for field, value in field_to_value.items()]
            return f'{{{", ".join(params)}}}'

        params = ', '.join(field_to_value.values())
        return f'[{params}]'

    @classmethod
    def dump_from_named_tuple_untyped(cls, tp: TypeInfo, extras: Extras):
        as_dict = extras['config'].namedtuple_as_dict
        return f'{tp.v()}._asdict()' if as_dict else f'list({tp.v()})'

    @classmethod
    def _build_dict_comp(cls, tp, v, i_next, k_next, v_next, kt, vt, extras):
        tp_k_next = tp.replace(origin=kt, i=i_next, prefix='k', index=None, val_name=None)
        string_k = cls.dump_dispatcher_for_annotation(tp_k_next, extras)

        tp_v_next = tp.replace(origin=vt, i=i_next, prefix='v', index=None, val_name=None)
        string_v = cls.dump_dispatcher_for_annotation(tp_v_next, extras)

        if k_next == string_k and v_next == string_v:
            # Easy path; shallow copy
            return f'{v}.copy()'

        return f'{{{string_k}: {string_v} for {k_next}, {v_next} in {v}.items()}}'

    @classmethod
    def dump_from_dict(cls, tp: TypeInfo, extras: Extras):
        try:
            kt, vt = tp.args
        except ValueError:
            # Annotated without two arguments,
            # e.g. like `dict[str]` or `dict`
            return f'{tp.v()}.copy()'

        v, k_next, v_next, i_next = tp.v_and_next_k_v()

        return cls._build_dict_comp(
            tp, v, i_next, k_next, v_next, kt, vt, extras)

    dump_from_defaultdict = dump_from_dict

    @classmethod
    def dump_from_typed_dict(cls, tp: TypeInfo, extras: Extras):
        req_keys, opt_keys = get_keys_for_typed_dict(tp.origin)
        if opt_keys:
            return cls._dump_from_typed_dict_fn(tp, extras)
        ann = tp.origin.__annotations__

        dict_body = ', '.join(
            f"""{name!r}: {
            cls.dump_dispatcher_for_annotation(
                tp.replace(origin=ann.get(name, Any), index=repr(name)),
                extras,
            )
            }"""
            for name in req_keys
        )

        return f'{{{dict_body}}}'

    @classmethod
    @setup_recursive_safe_function(prefix='dump')
    def _dump_from_typed_dict_fn(cls, tp: TypeInfo, extras: Extras):
        fn_gen = extras['fn_gen']
        req_keys, opt_keys = get_keys_for_typed_dict(tp.origin)
        td_annotations = tp.origin.__annotations__

        v = tp.v_for_def()
        result_list = []

        # Set required keys for the `TypedDict`
        for k in req_keys:
            field_tp = td_annotations[k]
            field_name = repr(k)
            string = cls.dump_dispatcher_for_annotation(
                tp.replace(origin=field_tp,
                           index=field_name,
                           val_name=None), extras)

            result_list.append(f'{field_name}: {string}')

        fn_gen.add_lines('result = {',
                         *(f'  {r},' for r in result_list),
                         '}')

        # Set optional keys for the `TypedDict` (if they exist)
        next_i = tp.i + 1
        new_tp = tp.replace(i=next_i, index=None, val_name=None)
        v_next = new_tp.v()

        for k in opt_keys:
            field_tp = td_annotations[k]
            field_name = repr(k)
            string = cls.dump_dispatcher_for_annotation(
                new_tp.replace(origin=field_tp), extras)
            with fn_gen.if_(f'({v_next} := {v}.get({field_name}, MISSING)) is not MISSING'):
                fn_gen.add_line(f'result[{field_name}] = {string}')
        fn_gen.add_line('return result')

    @classmethod
    @setup_recursive_safe_function_for_generic(prefix='dump', per_class_cache=True)
    def dump_from_union(cls, tp: TypeInfo, extras: Extras):
        fn_gen = extras['fn_gen']
        config = extras['config']
        actual_cls = extras['cls']
        _locals = extras['locals']

        args = tp.args
        field_i = tp.field_i
        i = tp.i
        v = tp.v_for_def()

        tag_key = config.tag_key or TAG
        auto_assign_tags = config.auto_assign_tags
        leaf_handling_as_subclass = config.leaf_handling == 'issubclass'

        in_optional = NoneType in args

        _locals['fields'] = args
        _locals['tag_key'] = tag_key

        leaf_types = []
        try_parse_lines = []
        dataclass_and_line = []
        has_dataclass = False

        for possible_tp in args:

            possible_tp = eval_forward_ref_if_needed(possible_tp, actual_cls)

            tp_new = TypeInfo(possible_tp, field_i=field_i, i=i)
            tp_new.in_optional = in_optional

            if _type_returns_value_unchanged(
                possible_tp, leaf_handling_as_subclass):
                leaf_types.append(possible_tp)

            # if num_leaf_types_no_bytes > 0:
            #     fn_gen.add_line(f'return {v}')

            elif is_dataclass(possible_tp):
                # we see a dataclass in `Union` declaration
                has_dataclass = True
                string = cls.dump_dispatcher_for_annotation(tp_new, extras)

                meta = get_meta(possible_tp)
                cls_name = possible_tp.__name__

                tag = meta.tag
                assign_tags_to_cls = auto_assign_tags or meta.auto_assign_tags
                if assign_tags_to_cls and not tag:
                    tag = cls_name
                    # We don't want to mutate the base Meta class here
                    if meta is AbstractMeta:
                        from ._meta_cache import create_meta
                        create_meta(possible_tp, cls_name, tag=tag)
                    else:
                        meta.tag = cls_name

                if tag:
                    dataclass_and_line.append(
                        (possible_tp, cls_name, tag,
                         f'result = {string}; result[tag_key] = {tag!r}; return result'))
                else:
                    dataclass_and_line.append(
                        (possible_tp, cls_name, tag,
                         f'return {string}'))

            else:
                try_parse_lines.append(
                    cls.dump_dispatcher_for_annotation(tp_new, extras))

        fn_gen.add_line(f't = {v}.__class__')

        if leaf_types:
            # a good heuristic: use tuples for smaller unions, else frozenset
            container = tuple if len(leaf_types) <= 6 else frozenset
            _locals['leaf_types'] = container(leaf_types)
            leaf_type_names = ', '.join(getattr(t, '__name__', None) or str(t)
                                        for t in leaf_types)
            with fn_gen.if_('t in leaf_types', comment=f'{{{leaf_type_names}}}'):
                fn_gen.add_line(f'return {v}')

        if has_dataclass:

            for field_i, (dataclass, _name, tag, line) in enumerate(dataclass_and_line, start=1):
                cls_name = TypeInfo(dataclass).type_name(extras)
                with fn_gen.if_(f't is {cls_name}', comment=f'{tag!r}' if tag else ''):
                    fn_gen.add_line(line)

        for string in try_parse_lines:
            with fn_gen.try_():
                fn_gen.add_line(f'return {string}')
            with fn_gen.except_(Exception):
                fn_gen.add_line('pass')

        # Invalid type for Union
        fn_gen.add_line("raise ParseError("
                        "TypeError('Object was not in any of Union types'),"
                        f"{v},fields,'dump',"
                        "tag_key=tag_key"
                        ")")

    @staticmethod
    def dump_from_decimal(tp: TypeInfo, extras: Extras):
        return f'str({tp.v()})'

    @staticmethod
    def dump_from_path(tp: TypeInfo, extras: Extras):
        return f'str({tp.v()})'

    @classmethod
    def dump_from_date(cls, tp: TypeInfo, extras: Extras):
        o = tp.v()

        if extras['config'].dump_date_time_as is DateTimeTo.TIMESTAMP:
            tp.ensure_in_locals(extras, datetime, UTC=UTC)
            return f'int(datetime((v0 := {o}).year, v0.month, v0.day, tzinfo=UTC).timestamp())'

        return f'{o}.isoformat()'

    @classmethod
    def dump_from_datetime(cls, tp: TypeInfo, extras: Extras):
        o = tp.v()
        config = extras['config']

        if config.dump_date_time_as is DateTimeTo.TIMESTAMP:
            naive_tz = config.assume_naive_datetime_tz

            if naive_tz is None:
                def raise_naive():
                    raise ValueError('Naive datetime has no timezone; '
                                     'set assume_naive_datetime_tz to '
                                     'define how it should be interpreted.')

                tp.ensure_in_locals(extras, raise_naive, ZERO=ZERO)
                return f'raise_naive() if (v0 := {o}).tzinfo is None else int(v0.timestamp()) if v0.utcoffset() == ZERO else int(v0.astimezone(UTC).timestamp())'

            else:
                tp.ensure_in_locals(extras, datetime_to_timestamp, assume_naive_tz=naive_tz)
                return f'datetime_to_timestamp({o}, assume_naive_tz)'

        # Safe: source is datetime.isoformat(); '+00:00' only appears for UTC
        return f"{o}.isoformat().replace('+00:00', 'Z', 1)"

    @staticmethod
    def dump_from_time(tp: TypeInfo, _extras: Extras):
        o = tp.v()
        # Safe: source is time.isoformat(); '+00:00' only appears for UTC
        return f"{o}.isoformat().replace('+00:00', 'Z', 1)"

    @staticmethod
    def dump_from_timedelta(tp: TypeInfo, extras: Extras):
        return f'str({tp.v()})'

    @staticmethod
    @setup_recursive_safe_function(
        prefix='dump',
        fn_name=f'__{PACKAGE_NAME}_to_dict_{{cls_name}}__')
    def dump_from_dataclass(tp: TypeInfo, extras: Extras):
        dump_func_for_dataclass(tp.origin, extras)

    @classmethod
    def dump_dispatcher_for_annotation(cls,
                                       tp,
                                       extras):

        hooks = cls.__HOOKS__
        config = extras['config']
        type_hooks = config.type_to_dump_hook
        leaf_handling_as_subclass = config.leaf_handling == 'issubclass'

        # type_ann = tp.origin
        type_ann = eval_forward_ref_if_needed(tp.origin, extras['cls'])

        origin = get_origin_v2(type_ann)
        name = getattr(origin, '__name__', origin)
        args = None

        if is_annotated(type_ann):
            # Given `Annotated[T, ...]`, we only need `T`
            type_ann, *field_extras = get_args(type_ann)
            type_ann = eval_forward_ref_if_needed(type_ann, extras['cls'])
            origin = get_origin_v2(type_ann)
            name = getattr(origin, '__name__', origin)
            # Check for Custom Patterns for date / time / datetime
            for extra in field_extras:
                if getattr(extra, '__dcw_pattern__', False):
                    extras['pattern'] = extra

        elif is_typed_dict_type_qualifier(origin):
            # Given `Required[T]` or `NotRequired[T]`, we only need `T`
            type_ann = get_args(type_ann)[0]
            origin = get_origin_v2(type_ann)
            name = getattr(origin, '__name__', origin)

        # TypeAliasType: Type aliases are created through
        # the `type` statement
        if (value := getattr(origin, '__value__', None)) is not None:
            type_ann = value
            origin = get_origin_v2(type_ann)
            name = getattr(origin, '__name__', origin)

        # `LiteralString` enforces stricter rules at
        # type-checking but behaves like `str` at runtime.
        # TODO maybe add `load_to_literal_string`
        if origin is PyLiteralString:
            dump_hook = cls.dump_from_str
            origin = str
            name = 'str'

        # -> Atomic, immutable types which don't require
        #    any iterative / recursive handling.
        elif origin in LEAF_TYPES or (
            leaf_handling_as_subclass
            and is_subclass_safe(origin, LEAF_TYPES)):
            dump_hook = hooks.get(origin)

        elif (type_hooks is not None
              and (hook_info := type_hooks.get(origin)) is not None):
            mode, dump_hook = hook_info

            if mode == 'runtime':
                fn_name, = tp.ensure_in_locals(extras, dump_hook)
                return f'{fn_name}({tp.v()})'

            try:
                args = get_args(type_ann)
            except ValueError:
                args = Any,

        elif (dump_hook := hooks.get(origin)) is not None:
            try:
                args = get_args(type_ann)
            except ValueError:
                args = Any,

        # -> Union[x]
        elif is_union(origin):
            args = get_args(type_ann)

            # all args in `Union[...]` are simple types
            if _all_return_value_unchanged(args, leaf_handling_as_subclass):
                return tp.v()

            dump_hook = cls.dump_from_union

            # Special case for Optional[x], which is actually Union[x, None]
            if len(args) == 2 and NoneType in args:
                origin = args[0]

                if tp.val_name:
                    val_name = 'v0'
                    o = f'({val_name} := {tp.v()})'
                else:
                    val_name = None
                    o = tp.v()

                new_tp = tp.replace(origin=origin, args=None, name=None, val_name=val_name)
                new_tp.in_optional = True

                string = cls.dump_dispatcher_for_annotation(new_tp, extras)

                return f'None if {o} is None else {string}'

        # -> Literal[X, Y, ...]
        elif origin is Literal:
            dump_hook = cls.dump_from_literal
            args = get_args(type_ann)

        # https://stackoverflow.com/questions/76520264/dataclasswizard-after-upgrading-to-python3-11-is-not-working-as-expected
        elif origin is Any:
            dump_hook = cls.dump_fallback

        elif is_subclass_safe(origin, tuple) and hasattr(origin, '_fields'):

            if getattr(origin, '__annotations__', None):
                # Annotated as a `typing.NamedTuple` subtype
                dump_hook = cls.dump_from_named_tuple
            else:
                # Annotated as a `collections.namedtuple` subtype
                dump_hook = cls.dump_from_named_tuple_untyped

        elif is_typed_dict(origin):
            dump_hook = cls.dump_from_typed_dict

        elif is_dataclass(origin):
            # return a dynamically generated `asdict`
            # for the `cls` (base_type)
            dump_hook = cls.dump_from_dataclass

        elif is_subclass_safe(origin, Enum):
            dump_hook = cls.dump_from_enum

        elif origin in (abc.Sequence, abc.MutableSequence, abc.Collection):
            if origin is abc.Sequence:
                dump_hook = cls.dump_from_tuple
                # desired (non-generic) origin type
                name = 'tuple'
                origin = tuple
                # Re-map type arguments to variadic tuple format,
                # e.g. `Sequence[int]` -> `tuple[int, ...]`
                try:
                    args = (get_args(type_ann)[0], ...)
                except (IndexError, ValueError):
                    args = Any,
            else:
                dump_hook = cls.dump_from_iterable
                # desired (non-generic) origin type
                name = 'list'
                origin = list
                # Get type arguments, e.g. `Sequence[int]` -> `int`
                try:
                    args = get_args(type_ann)
                except ValueError:
                    args = Any,

        elif getattr(origin, '__dcw_pattern__', False):
            __base__ = origin.base

            if issubclass(__base__, datetime):
                dump_hook = cls.dump_from_datetime
                origin = datetime
            elif issubclass(__base__, date):
                dump_hook = cls.dump_from_date
                origin = date
            elif issubclass(__base__, time):
                dump_hook = cls.dump_from_time
                origin = time

        elif origin is None:
            dump_hook = cls.dump_from_none

        else:

            # TODO everything should use `get_origin_v2`
            try:
                args = get_args(type_ann)
            except ValueError:
                args = Any,

        if dump_hook is None:
            # TODO END
            for t in hooks:
                if (not leaf_handling_as_subclass) and (t in LEAF_TYPES):
                    continue
                if issubclass(origin, t):
                    dump_hook = hooks[t]
                    break

        tp.origin = origin
        tp.args = args
        tp.name = name

        if dump_hook is not None:
            result = dump_hook(tp, extras)
            return result

        # No matching hook is found for the type.
        # TODO do we want to add a `Meta` field to not raise
        #  an error but perform a default action?
        err = TypeError('Provided type is not currently supported.')
        pe = ParseError(
            err, origin, type_ann, 'dump',
            resolution=f'Register a dump hook for {ParseError.name(origin)} '
                       f'(`register_type` / `Meta.type_to_dump_hook`).',
            unsupported_type=origin
        )
        raise pe from None


def get_default_dump_hooks(dumper):
    return {
        # Technically a complex type, however check this
        # first, since `StrEnum` and `IntEnum` are subclasses
        # of `str` and `int`
        Enum: dumper.dump_from_enum,
        # Simple types
        str: dumper.dump_from_str,
        float: dumper.dump_from_float,
        bool: dumper.dump_from_bool,
        int: dumper.dump_from_int,
        bytes: dumper.dump_from_bytes,
        bytearray: dumper.dump_from_bytearray,
        NoneType: dumper.dump_from_none,
        # Complex types
        UUID: dumper.dump_from_uuid,
        set: dumper.dump_from_iterable,
        frozenset: dumper.dump_from_iterable,
        deque: dumper.dump_from_iterable,
        list: dumper.dump_from_iterable,
        tuple: dumper.dump_from_tuple,
        defaultdict: dumper.dump_from_defaultdict,
        dict: dumper.dump_from_dict,
        Decimal: dumper.dump_from_decimal,
        Path: dumper.dump_from_path,
        # Dates and times
        datetime: dumper.dump_from_datetime,
        time: dumper.dump_from_time,
        date: dumper.dump_from_date,
        timedelta: dumper.dump_from_timedelta,
    }


def setup_default_dumper(cls=DumpMixin):
    """
    Setup the default type hooks to use when converting
    a `dataclass` instance to a `str` (json) or a
    Python `dict` object.

    Note: `cls` must be :class:`DumpMixIn` or a sub-class of it.
    """
    if '__HOOKS__' in cls.__dict__:
        return

    parent_hooks = getattr(cls, '__HOOKS__', None)

    hooks = get_default_dump_hooks(cls)
    if parent_hooks:
        hooks |= parent_hooks  # parent / custom wins

    cls.__HOOKS__ = hooks


def check_and_raise_missing_fields(
    _locals, o, cls, fields: tuple[Field, ...]):
    missing_fields = [f.name for f in fields
                      if f.init
                      and f'__{f.name}' not in _locals
                      and (f.default is MISSING
                           and f.default_factory is MISSING)]

    missing_keys = [resolve_dataclass_field_to_alias_for_dump(cls)[field]
                    for field in missing_fields]

    raise MissingFields(
        None, o, cls, fields, None, missing_fields,
        missing_keys
    ) from None


def dump_func_for_dataclass(
    cls: type,
    extras: Extras | None = None,
    dumper_cls=DumpMixin,
    base_meta_cls: type = AbstractMeta,
) -> Callable[[T], JSONObject] | str:
    # TODO dynamically generate for multiple nested classes at once

    # Tuple describing the fields of this dataclass.
    cls_fields = dataclass_fields(cls)

    cls_fields_list = list(cls_fields)
    cls_field_names = dataclass_field_names(cls)

    # Get the dumper for the class, or create a new one as needed.
    cls_dumper = get_dumper(cls, base_cls=dumper_cls)

    cls_name = cls.__name__

    fn_name = f'__{PACKAGE_NAME}_to_dict_{cls_name}__'

    # Get the meta config for the class, or the default config otherwise.
    meta = get_meta(cls, base_meta_cls)

    if extras is None:  # we are being run for the main dataclass
        is_main_class = True

        # If the `recursive` flag is enabled and a Meta config is provided,
        # apply the Meta recursively to any nested classes.
        #
        # Else, just use the base `AbstractMeta`.
        config: META = meta if meta.recursive else base_meta_cls

        # Initialize the FuncBuilder
        fn_gen = FunctionBuilder()

        new_locals = {
            'cls': cls,
            'fields': cls_fields,
        }

        # noinspection PyTypeChecker
        extras: Extras = {
            'config': config,
            'cls': cls,
            'cls_name': cls_name,
            'locals': new_locals,
            'recursion_guard': {cls: fn_name},
            'fn_gen': fn_gen,
        }

        _globals = {
            'MISSING': MISSING,
            'ParseError': ParseError,
            'raise_missing_fields': check_and_raise_missing_fields,
            're_raise': re_raise,
        }

    # we are being run for a nested dataclass
    else:
        is_main_class = False

        # config for nested dataclasses
        config = extras['config']

        # Initialize the FuncBuilder
        fn_gen = extras['fn_gen']

        if config is not base_meta_cls:
            # we want to apply the meta config from the main dataclass
            # recursively.
            meta = meta | config
            meta.bind_to(cls, is_default=False)

        new_locals = extras['locals']
        new_locals['fields'] = cls_fields

        # TODO need a way to auto-magically do this
        extras['cls'] = cls
        extras['cls_name'] = cls_name

    key_case: KeyCase | None = cls_dumper.transform_dataclass_field

    # TODO decide if different logic is needed for `AUTO` case
    if key_case is KeyCase.AUTO:
        key_case = None

    # A cached mapping of each dataclass field to the resolved key name in a
    # JSON or dictionary object; useful so we don't need to do a case
    # transformation (via regex) each time.
    field_to_alias = resolve_dataclass_field_to_alias_for_dump(cls)
    check_aliases = True if field_to_alias else False

    field_to_path = DATACLASS_FIELD_TO_ALIAS_PATH_FOR_DUMP[cls]
    has_paths = True if field_to_path else False

    # A cached mapping of dataclass field name to its default value, either
    # via a `default` or `default_factory` argument.
    # FIXME get from functions instead
    has_defaults = SEEN_DEFAULT[cls]

    # A cached mapping of dataclass field name to its SkipIf condition.
    field_to_skip_if = dataclass_field_to_skip_if(cls)

    skip_if_condition = get_skip_if_condition(
        meta.skip_if, new_locals, '_skip_value')
    skip_defaults_if_condition = get_skip_if_condition(
        meta.skip_defaults_if, new_locals, '_skip_defaults_value')

    skip_defaults = True if meta.skip_defaults else False
    skip_if = True if field_to_skip_if or skip_if_condition else False

    catch_all_name: str | None = field_to_alias.pop(CATCH_ALL, None)
    has_catch_all = catch_all_name is not None

    if has_catch_all:
        catch_all_name_stripped = catch_all_name.rstrip('?')
        catch_all_idx = cls_field_names.index(catch_all_name_stripped)
        # remove catch all field from list, so we don't iterate over it
        # noinspection PyTypeChecker
        catch_all_field = cls_fields_list.pop(catch_all_idx)
    else:
        catch_all_name_stripped = catch_all_field = None

    cls_name = cls.__name__

    with fn_gen.function(
        fn_name, [
            'o',
            'dict_factory=dict',
            "exclude:'list[str]|None'=None",
            f'skip_defaults:bool={skip_defaults}',
        ], MISSING, new_locals):

        if (_pre_to_dict := getattr(cls, '_pre_to_dict', None)) is not None:
            new_locals['__pre_to_dict__'] = _pre_to_dict
            fn_gen.add_line('o = __pre_to_dict__(o)')

        # Need to create a separate dictionary to copy over the constructor
        # args, as we don't want to mutate the original dictionary object.
        if has_defaults:
            fn_gen.add_line('add_defaults = not skip_defaults')

        if has_paths:
            new_locals['NestedDict'] = NestedDict
            fn_gen.add_line('paths = NestedDict()')

        required_field_assigns = []
        default_assigns = []
        path_assigns = []
        name_to_skip_condition = {}

        if cls_fields_list:

            with fn_gen.try_():

                for i, f in enumerate(cls_fields_list):
                    name = f.name
                    has_skip_if = False
                    # TODO: This is if we want to check if field is in `exclude`
                    #  (not huge performance gain)
                    # skip_field = f'_skip_{i}'
                    default_value = default_compare_expr(f, new_locals, f'_default_{i}')
                    has_default = default_value is not None

                    # Check for Field Aliases + Paths
                    # NOTE: `key` is used later, so we need to capture it.
                    if check_aliases and (key := field_to_alias.get(name)) is not None:
                        # special case: skip serialization for field, e.g. `Alias(..., skip=True)`
                        if key is ExplicitNull:
                            continue

                    elif key_case is None:
                        key = name

                    else:
                        key = key_case(name)

                    # If field has an explicit `SkipIf` condition
                    if skip_if:
                        has_skip_if = True
                        if (_skip_condition := field_to_skip_if.get(name)) is not None:
                            _skip_if = get_skip_if_condition(
                                _skip_condition, new_locals, condition_i=i)
                            _final_skip_if = finalize_skip_if(_skip_condition, '{}', _skip_if)
                            name_to_skip_condition[name] = f'not ({_final_skip_if})'

                        # If Meta `skip_if` has a value
                        elif skip_if_condition:
                            _final_skip_if = finalize_skip_if(
                                meta.skip_if, '{}', skip_if_condition)
                            name_to_skip_condition[name] = f'not ({_final_skip_if})'

                        # # Else, proceed as normal
                        # else:
                        #     field_assignments.append(f"if not {skip_field}:")

                    # A dataclass field which specifies a "JSON Path".
                    if has_paths and (
                        path := field_to_path.get(name)
                    ) is not None:  # AliasPath(...)
                        lvalue = f"paths{''.join(f'[{p!r}]' for p in path)}"
                        if has_default:
                            string = generate_field_code(cls_dumper, extras, f, i)
                            default_assigns.append((name, key, default_value, lvalue, string))
                        else:
                            var_name = 'v1' if has_skip_if else f'o.{name}'
                            string = generate_field_code(cls_dumper, extras, f, i, var_name)
                            path_assigns.append((name, f"{lvalue} = {string}"))

                        continue

                    if has_default:
                        string = generate_field_code(cls_dumper, extras, f, i)
                        lvalue = f'result[{key!r}]'
                        default_assigns.append((name, key, default_value, lvalue, string))
                    else:
                        # TODO confirm this is ok
                        # vars_for_fields.append(f'{name}={var}')
                        if has_skip_if:
                            string = generate_field_code(cls_dumper, extras, f, i, 'v1')
                            lvalue = f'result[{key!r}]'
                            default_assigns.append((name, ExplicitNull, None, lvalue, string))
                        else:
                            string = generate_field_code(cls_dumper, extras, f, i, f'o.{name}')
                            required_field_assigns.append((name, key, string))

                # Add assignments for `AliasPath(...)`
                for (name, line) in path_assigns:
                    if (condition := name_to_skip_condition.get(name)) is not None:
                        fn_gen.add_line(f'v1 = o.{name}')
                        with fn_gen.if_(condition.format('v1')):
                            fn_gen.add_line(line)
                    else:
                        fn_gen.add_line(line)

                # Add required dataclass field assignments
                fn_gen.add_line('result = {')
                for (_, key, string) in required_field_assigns:
                    fn_gen.add_line(f'  {key!r}: {string},')
                fn_gen.add_line('}')

                # Add default (optional) dataclass field assignments
                for (name, key, default_name, lvalue, rvalue) in default_assigns:
                    var_name = 'v1'
                    if rvalue == var_name:  # and default_name is not ExplicitNull:
                        var_name = rvalue = f'o.{name}'
                    else:
                        fn_gen.add_line(f'{var_name} = o.{name}')

                    line = f'{lvalue} = {rvalue}'
                    def_condition = f'add_defaults or {var_name} != {default_name}'

                    if skip_defaults_if_condition and key is not ExplicitNull:
                        _final_skip_if = finalize_skip_if(
                            meta.skip_defaults_if, var_name, skip_defaults_if_condition)
                        # TODO missing skip individual condition!!
                        with fn_gen.if_(
                            f'(add_defaults or {var_name} != {default_name}) '
                            f'and not ({_final_skip_if})'):
                            fn_gen.add_line(line)

                    elif (condition := name_to_skip_condition.get(name)) is not None:
                        condition = condition.format(var_name)
                        if default_name is None:  # Required field with skip condition
                            with fn_gen.if_(condition):
                                fn_gen.add_line(line)
                        else:
                            with fn_gen.if_(
                                f'(add_defaults or {var_name} != {default_name}) '
                                f'and {condition}'):
                                fn_gen.add_line(line)

                    else:
                        with fn_gen.if_(def_condition):
                            fn_gen.add_line(line)

            # create a broad `except Exception` block, as we will be
            # re-raising all exception(s) as a custom `ParseError`.
            with fn_gen.except_(Exception, 'e', ParseError):
                fn_gen.add_line("re_raise(e, cls, o, fields, '<UNK>', locals().get('v1'))")

        else:
            fn_gen.add_line('result = {}')

        if has_catch_all:
            # noinspection PyUnresolvedReferences,PyProtectedMember
            # TODO
            from dataclasses import (
                _asdict_inner as __dataclasses_asdict_inner__,
            )

            if (default_value := default_compare_expr(
                    catch_all_field,
                    new_locals,
                f'_default_{len(cls_fields_list)}')) is not None:
                condition = f"(v1 := o.{catch_all_name_stripped}) != {default_value}"

            else:
                condition = f'v1 := o.{catch_all_name_stripped}'

            with fn_gen.if_(condition):
                with fn_gen.for_("k, v in v1.items()"):
                    fn_gen.globals['__asdict_inner__'] = __dataclasses_asdict_inner__
                    fn_gen.add_line('result[k] = __asdict_inner__(v,dict_factory)')

        with fn_gen.if_('exclude'):
            with fn_gen.for_('k in exclude'):
                fn_gen.add_line('result.pop(k, None)')

        if has_paths:
            fn_gen.add_line('result.update(paths)')

        # Now pass the arguments to the dict_factory method, and return
        # the new dict_factory instance.
        fn_gen.add_line('return result if dict_factory is dict else dict_factory(result)')

    # Save the dump function for the main dataclass, so we don't need to run
    # this logic each time.
    if is_main_class:
        # noinspection PyUnboundLocalVariable
        functions = fn_gen.create_functions(_globals)

        cls_todict = functions[fn_name]

        # Check if the class has a `to_dict`, and it's
        # a class method bound to `todict`.
        if getattr(cls, 'to_dict', None) is asdict:
            LOG.debug("setattr(%s, 'to_dict', %s)", cls_name, fn_name)
            # Marker reserved for future detection/debugging of specialized dumpers.
            # setattr(cls_todict, _SPECIALIZED_TO_DICT, True)
            # safe to specialize only when user didn't define it on cls
            set_new_attribute(cls, 'to_dict', cls_todict, force=True)

        set_new_attribute(
            cls, '__dataclass_wizard_to_dict__', cls_todict)
        LOG.debug(
            "setattr(%s, '__%s_to_dict__', %s)",
            cls_name, PACKAGE_NAME, fn_name)

        return cls_todict


def generate_field_code(cls_dumper: DumpMixin,
                        extras: Extras,
                        field: Field,
                        field_i: int,
                        var_name=None) -> str | TypeInfo:
    cls = extras['cls']
    field_type = field.type = eval_forward_ref_if_needed(field.type, cls)

    try:
        return cls_dumper.dump_dispatcher_for_annotation(
            TypeInfo(field_type, field_i=field_i, val_name=var_name), extras
        )

    except ParseError as pe:
        pe.class_name = cls
        pe.field_name = field.name
        raise pe from None


def re_raise(e, cls, o, fields, field, value):
    # If the object `o` is None, then raise an error with
    # the relevant info included.
    if o is None:
        raise MissingData(cls) from None

    add_fields = True
    if type(e) is not ParseError:
        if isinstance(e, JSONWizardError):
            add_fields = False
        else:
            tp = getattr(next((f for f in fields if f.name == field), None), 'type', Any)
            e = ParseError(e, value, tp, 'dump')

    # If field name is missing or not known, make a "best effort"
    # to resolve it.
    if field == '<UNK>' and cls and fields:
        if len(names := [f.name for f in fields
                          if getattr(o, f.name, MISSING) == e.obj]) == 1:
            field = e.field_name = names[0]

    # We run into a parsing error while dumping the field value;
    # Add additional info on the Exception object before re-raising it.
    #
    # First confirm these values are not already set by an
    # inner dataclass. If so, it likely makes it easier to
    # debug the cause. Note that this should already be
    # handled by the `setter` methods.
    if add_fields:
        e.class_name, e.fields, e.field_name, e.json_object = cls, fields, field, repr(o)
    else:
        e.class_name, e.field_name, e.json_object = cls, field, repr(o)

    raise e from None


def get_dumper(class_or_instance=None, create=True,
               base_cls: T = DumpMixin) -> type[T]:
    """
    Get the dumper for the class, using the following logic:

        * Return the class if it's already a sub-class of :class:`DumpMixin`
        * If `create` is enabled (which is the default), a new sub-class of
          :class:`DumpMixin` for the class will be generated and cached on the
          initial run.
        * Otherwise, we will return the base loader, :class:`DumpMixin`, which
          can potentially be shared by more than one dataclass.

    """
    # TODO
    try:
        return CLASS_TO_DUMPER[class_or_instance]

    except KeyError:
        # TODO figure out type errors

        if hasattr(class_or_instance, _HOOKS):
            return set_class_dumper(
                CLASS_TO_DUMPER, class_or_instance, class_or_instance)

        elif create:
            cls_loader = create_new_class(class_or_instance, (base_cls, ))
            return set_class_dumper(
                CLASS_TO_DUMPER, class_or_instance, cls_loader)

        return set_class_dumper(
            CLASS_TO_DUMPER, class_or_instance, base_cls)


def asdict(o: T,
           *, cls=None,
           dict_factory=dict,
           exclude: Collection[str] | None = None,
           **kwargs) -> JSONObject:
    # noinspection PyUnresolvedReferences
    """Return the fields of a dataclass instance as a new dictionary mapping
    field names to field values.

    Example usage:

      @dataclass
      class C:
          x: int
          y: int

      c = C(1, 2)
      assert asdict(c) == {'x': 1, 'y': 2}

    When directly invoking this function, an optional Meta configuration for
    the dataclass can be specified via ``DumpMeta``; by default, this will
    apply recursively to any nested dataclasses. Here's a sample usage of this
    below::

        >>> DumpMeta(key_transform='CAMEL').bind_to(MyClass)
        >>> asdict(MyClass(my_str="value"))

    If given, 'dict_factory' will be used instead of built-in dict.
    The function applies recursively to field values that are
    dataclass instances. This will also look into built-in containers:
    tuples, lists, and dicts.
    """
    cls = cls or type(o)

    try:
        return cls.__dataclass_wizard_to_dict__(
            o, dict_factory, exclude, **kwargs)

    except (AttributeError, TypeError):
        fn = dump_func_for_dataclass(cls)
        cls.__dataclass_wizard_to_dict__ = fn  # explicit cache
        return fn(
            o, dict_factory, exclude, **kwargs)



from __future__ import annotations

from weakref import WeakKeyDictionary

from ._bases import AbstractMeta
from ._type_def import META

META_BY_DATACLASS = WeakKeyDictionary()

# Injected at runtime by bases_meta.py
BASE_META_CLS = None


def set_base_meta_cls(base_meta_cls):
    global BASE_META_CLS
    BASE_META_CLS = base_meta_cls


def get_meta(cls, base_cls=AbstractMeta):
    """
    Retrieves the Meta config for the :class:`AbstractJSONWizard` subclass.

    This config is set when the inner :class:`Meta` is sub-classed.
    """
    return META_BY_DATACLASS.get(cls, base_cls)


def create_meta(cls, cls_name=None, **kwargs):
    """
    Create a Meta subclass for `cls` and store it in META_BY_DATACLASS.
    Requires `set_base_meta_cls` to have been called.

    WARNING: Only use if the Meta config is undefined,
      e.g. `get_meta` for the `cls` returns `base_cls`.
    """
    base = BASE_META_CLS
    if base is None:
        # Fail fast with a helpful error instead of mysterious circular-import states.
        raise RuntimeError(
            'Base meta class not initialized. '
            'Expected set_base_meta_cls(BaseJSONWizardMeta) to be called during import.'
        )

    cls_dict = {'__slots__': (), **kwargs}

    meta: META = type(   # type: ignore
        f'{(cls_name or cls.__name__)}Meta',
        (base, ),
        cls_dict,
    )

    META_BY_DATACLASS[cls] = meta
    return meta



from typing import Any

from .._type_def import FREF

__all__ = ['is_union',
           'get_origin',
           'get_origin_v2',
           'is_typed_dict_type_qualifier',
           'get_args',
           'get_keys_for_typed_dict',
           'is_typed_dict',
           'is_generic',
           'is_annotated',
           'eval_forward_ref',
           'eval_forward_ref_if_needed']

def get_args(tp: Any) -> tuple[Any, ...]: ...
def get_keys_for_typed_dict(cls): ...
def is_typed_dict_type_qualifier(cls) -> bool: ...
def is_union(cls) -> bool: ...
def get_origin_v2(cls): ...
def is_typed_dict(cls: type) -> bool: ...
def is_generic(cls): ...
def get_origin(cls, raise_: bool = ...): ...
def is_annotated(cls): ...
def eval_forward_ref(base_type: FREF, cls: type): ...
def eval_forward_ref_if_needed(base_type, base_cls: type): ...



import logging
from dataclasses import dataclass
from typing import Any

import pytest

from dataclasses_json import (dataclass_json, Undefined, CatchAll as CatchAllDJ)
from dataclass_wizard import (DataclassWizard, CatchAll as CatchAllWizard)


log = logging.getLogger(__name__)


@dataclass()
class DontCareAPIDump:
    endpoint: str
    data: dict[str, Any]


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclass()
class DontCareAPIDumpDJ(DontCareAPIDump):
    unknown_things: CatchAllDJ


class DontCareAPIDumpWizard(DontCareAPIDump, DataclassWizard):

    # `v1=True` is the default with `DataclassWizard`
    # class _(JSONWizard.Meta):
    #     v1 = True

    unknown_things: CatchAllWizard


# Fixtures for test data
@pytest.fixture(scope='session')
def data():
    return {"endpoint": "some_api_endpoint",
            "data": {"foo": 1, "bar": "2"},
            "undefined_field_name": [1, 2, 3]}


@pytest.fixture(scope='session')
def data_no_extras():
    return {"endpoint": "some_api_endpoint",
            "data": {"foo": 1, "bar": "2"}}


# Benchmark for deserialization (from_dict)
@pytest.mark.benchmark(group="deserialization")
def test_deserialize_wizard(benchmark, data):
    benchmark(lambda: DontCareAPIDumpWizard.from_dict(data))


@pytest.mark.benchmark(group="deserialization")
def test_deserialize_json(benchmark, data):
    benchmark(lambda: DontCareAPIDumpDJ.from_dict(data))


# Benchmark for deserialization with no extra data
@pytest.mark.benchmark(group="deserialization_no_extra_data")
def test_deserialize_wizard_no_extras(benchmark, data_no_extras):
    benchmark(lambda: DontCareAPIDumpWizard.from_dict(data_no_extras))


@pytest.mark.benchmark(group="deserialization_no_extra_data")
def test_deserialize_json_no_extras(benchmark, data_no_extras):
    benchmark(lambda: DontCareAPIDumpDJ.from_dict(data_no_extras))


# Benchmark for serialization (to_dict)
@pytest.mark.benchmark(group="serialization")
def test_serialize_wizard(benchmark, data):
    dump1 = DontCareAPIDumpWizard.from_dict(data)
    benchmark(lambda: dump1.to_dict())


@pytest.mark.benchmark(group="serialization")
def test_serialize_json(benchmark, data):
    dump2 = DontCareAPIDumpDJ.from_dict(data)
    benchmark(lambda: dump2.to_dict())


def test_validate(data, data_no_extras):
    dump1 = DontCareAPIDumpDJ.from_dict(data_no_extras)  # DontCareAPIDump(endpoint='some_api_endpoint', data={'foo': 1, 'bar': '2'})
    dump2 = DontCareAPIDumpWizard.from_dict(data_no_extras)  # DontCareAPIDump(endpoint='some_api_endpoint', data={'foo': 1, 'bar': '2'})

    assert dump1.endpoint == dump2.endpoint
    assert dump1.data == dump2.data
    assert dump1.unknown_things == dump2.unknown_things == {}

    expected = {'endpoint': 'some_api_endpoint', 'data': {'foo': 1, 'bar': '2'}}

    assert dump1.to_dict() == dump2.to_dict() == expected

    dump1 = DontCareAPIDumpDJ.from_dict(data)  # DontCareAPIDump(endpoint='some_api_endpoint', data={'foo': 1, 'bar': '2'})
    dump2 = DontCareAPIDumpWizard.from_dict(data)  # DontCareAPIDump(endpoint='some_api_endpoint', data={'foo': 1, 'bar': '2'})

    assert dump1.endpoint == dump2.endpoint
    assert dump1.data == dump2.data
    assert dump1.unknown_things == dump2.unknown_things

    expected = {'endpoint': 'some_api_endpoint', 'data': {'foo': 1, 'bar': '2'}, 'undefined_field_name': [1, 2, 3]}

    assert dump1.to_dict() == dump2.to_dict() == expected



from __future__ import annotations

import logging
from dataclasses import dataclass, field

from dataclass_wizard.v0 import property_wizard


log = logging.getLogger(__name__)


def test_property_wizard_with_public_property_and_field_with_or():
    """
    Using `property_wizard` when the dataclass has both a property and field
    name *without* a leading underscore, and using the OR ("|") operator,
    instead of the `typing.Union` usage.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        # The value of `wheels` here will be ignored, since `wheels` is simply
        # re-assigned on the following property definition.
        wheels: int | str = 4

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: int | str):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_unresolvable_forward_ref():
    """
    Using `property_wizard` when the annotated field for a property references
    a class or type that is not yet declared.
    """
    @dataclass
    class Car:
        spare_tires: int

    class Truck:
        ...

    globals().update(locals())

    @dataclass
    class Vehicle(metaclass=property_wizard):

        # The value of `cars` here will be ignored, since `cars` is simply
        # re-assigned on the following property definition.
        cars: list[Car] = field(default_factory=list)
        trucks: list[Truck] = field(default_factory=list)

        @property
        def cars(self) -> int:
            return self._cars

        @cars.setter
        def cars(self, cars: list[Car]):
            self._cars = cars * 2 if cars else cars


    v = Vehicle()
    log.debug(v)
    assert not v.cars
    # assert v.cars is None

    v = Vehicle([Car(1)])
    log.debug(v)
    assert v.cars == [Car(1), Car(1)], 'The constructor should use our ' \
                                       'setter method'

    v.cars = [Car(3)]
    assert v.cars == [Car(3), Car(3)], 'Expected assignment to use the ' \
                                       'setter method'



import json
from dataclasses import MISSING, Field
from datetime import date, datetime, time
from typing import (Collection, Callable,
                    Generic, Mapping, TypeAlias)
from typing import TypedDict, overload, Any, NotRequired

from .bases import META
from .decorators import cached_property
from .type_def import T, DT, Encoder, FileEncoder
from .utils.function_builder import FunctionBuilder
from .utils.object_path import PathPart, PathType


# Define a simple type (alias) for the `CatchAll` field
CatchAll: TypeAlias = Mapping | None

# Type for a string or a collection of strings.
_STR_COLLECTION: TypeAlias = str | Collection[str]


class Extras(TypedDict):
    """
    "Extra" config that can be used in the load / dump process.
    """
    config: NotRequired[META]
    cls: type
    cls_name: str
    fn_gen: FunctionBuilder
    locals: dict[str, Any]
    pattern: NotRequired[PatternedDT]


def json_key(*keys: str, all=False, dump=True):
    """
    Represents a mapping of one or more JSON key names for a dataclass field.

    This is only in *addition* to the default key transform; for example, a
    JSON key appearing as "myField", "MyField" or "my-field" will already map
    to a dataclass field "my_field" by default (assuming the key transform
    converts to snake case).

    The mapping to each JSON key name is case-sensitive, so passing "myfield"
    will not match a "myField" key in a JSON string or a Python dict object.

    :param keys: A list of one of more JSON keys to associate with the
      dataclass field.
    :param all: True to also associate the reverse mapping, i.e. from
      dataclass field to JSON key. If multiple JSON keys are passed in, it
      uses the first one provided in this case. This mapping is then used when
      `to_dict` or `to_json` is called, instead of the default key transform.
    :param dump: False to skip this field in the serialization process to
      JSON. By default, this field and its value is included.
    """
    ...


# noinspection PyPep8Naming
def KeyPath(keys: PathType | str, all: bool = True, dump: bool = True):
    """
    Represents a mapping of one or more "nested" key names in JSON
    for a dataclass field.

    This is only in *addition* to the default key transform; for example, a
    JSON key appearing as "myField", "MyField" or "my-field" will already map
    to a dataclass field "my_field" by default (assuming the key transform
    converts to snake case).

    The mapping to each JSON key name is case-sensitive, so passing "myfield"
    will not match a "myField" key in a JSON string or a Python dict object.

    :param keys: A list of one of more "nested" JSON keys to associate
      with the dataclass field.
    :param all: True to also associate the reverse mapping, i.e. from
      dataclass field to "nested" JSON key. If multiple JSON keys are passed in, it
      uses the first one provided in this case. This mapping is then used when
      `to_dict` or `to_json` is called, instead of the default key transform.
    :param dump: False to skip this field in the serialization process to
      JSON. By default, this field and its value is included.

    Example:

    >>> from typing import Annotated
    >>> my_str: Annotated[str, KeyPath('my."7".nested.path.-321')]
    >>> # where path.keys == ('my', '7', 'nested', 'path', -321)
    """
    ...


def env_field(keys: _STR_COLLECTION, *,
              all=False, dump=True,
              default=MISSING,
              default_factory: Callable[[], MISSING] = MISSING,
              init=True, repr=True,
              hash=None, compare=True, metadata=None):
    """
    This is a helper function that sets the same defaults for keyword
    arguments as the ``dataclasses.field`` function. It can be thought of as
    an alias to ``dataclasses.field(...)``, but one which also represents
    a mapping of one or more environment variable (env var) names to
    a dataclass field.

    This is only in *addition* to the default key transform; for example, an
    env var appearing as "myField", "MyField" or "my-field" will already map
    to a dataclass field "my_field" by default (assuming the key transform
    converts to snake case).

    `keys` is a string, or a collection (list, tuple, etc.) of strings. It
    represents one of more env vars to associate with the dataclass field.

    When `all` is passed as True (default is False), it will also associate
    the reverse mapping, i.e. from dataclass field to env var. If multiple
    env vars are passed in, it uses the first one provided in this case.
    This mapping is then used when ``to_dict`` or ``to_json`` is called,
    instead of the default key transform.

    When `dump` is passed as False (default is True), this field will be
    skipped, or excluded, in the serialization process to JSON.
    """
    ...


def json_field(keys: _STR_COLLECTION, *,
               all=False, dump=True,
               default=MISSING,
               default_factory: Callable[[], MISSING] = MISSING,
               init=True, repr=True,
               hash=None, compare=True, metadata=None):
    """
    This is a helper function that sets the same defaults for keyword
    arguments as the ``dataclasses.field`` function. It can be thought of as
    an alias to ``dataclasses.field(...)``, but one which also represents
    a mapping of one or more JSON key names to a dataclass field.

    This is only in *addition* to the default key transform; for example, a
    JSON key appearing as "myField", "MyField" or "my-field" will already map
    to a dataclass field "my_field" by default (assuming the key transform
    converts to snake case).

    The mapping to each JSON key name is case-sensitive, so passing "myfield"
    will not match a "myField" key in a JSON string or a Python dict object.

    `keys` is a string, or a collection (list, tuple, etc.) of strings. It
    represents one of more JSON keys to associate with the dataclass field.

    When `all` is passed as True (default is False), it will also associate
    the reverse mapping, i.e. from dataclass field to JSON key. If multiple
    JSON keys are passed in, it uses the first one provided in this case.
    This mapping is then used when ``to_dict`` or ``to_json`` is called,
    instead of the default key transform.

    When `dump` is passed as False (default is True), this field will be
    skipped, or excluded, in the serialization process to JSON.
    """
    ...


def path_field(keys: _STR_COLLECTION, *,
               all=True, dump=True,
               default=MISSING,
               default_factory: Callable[[], MISSING] = MISSING,
               init=True, repr=True,
               hash=None, compare=True, metadata=None):
    """
    Creates a dataclass field mapped to one or more nested JSON paths.

    This function is an alias for ``dataclasses.field(...)``, with additional
    logic for associating a field with one or more JSON key paths, including
    nested structures. It can be used to specify custom mappings between
    dataclass fields and complex, nested JSON key names.

    This mapping is **case-sensitive** and applies to the provided JSON keys
    or nested paths. For example, passing "myField" will not match "myfield"
    in JSON, and vice versa.

    `keys` represents one or more nested JSON keys (as strings or a collection of strings)
    to associate with the dataclass field. The keys can include paths like `a.b.c`
    or even more complex nested paths such as `a["nested"]["key"]`.

    Arguments:
        keys (_STR_COLLECTION): The JSON key(s) or nested path(s) to associate with the dataclass field.
        all (bool): If True (default), it also associates the reverse mapping
                    (from dataclass field to JSON path) for serialization.
                    This reverse mapping is used during `to_dict` or `to_json` instead
                    of the default key transform.
        dump (bool): If False (default is True), excludes this field from
                     serialization to JSON.
        default (Any): The default value for the field. Mutually exclusive with `default_factory`.
        default_factory (Callable[[], Any]): A callable to generate the default value.
                                             Mutually exclusive with `default`.
        init (bool): Include the field in the generated `__init__` method. Defaults to True.
        repr (bool): Include the field in the `__repr__` output. Defaults to True.
        hash (bool): Include the field in the `__hash__` method. Defaults to None.
        compare (bool): Include the field in comparison methods. Defaults to True.
        metadata (dict): Metadata to associate with the field. Defaults to None.

    Returns:
        JSONField: A dataclass field with logic for mapping to one or more nested JSON paths.

    Example:
        >>> from dataclasses import dataclass
        >>> @dataclass
        >>> class Example:
        >>>     my_str: str = path_field(['a.b.c.1', 'x.y["-1"].z'], default=42)
        >>> # Maps nested paths ('a', 'b', 'c', 1) and ('x', 'y', '-1', 'z')
        >>> # to the `my_str` attribute.
    """
    ...


def skip_if_field(condition: Condition, *,
                  default=MISSING,
                  default_factory: Callable[[], MISSING] = MISSING,
                  init=True, repr=True,
                  hash=None, compare=True, metadata=None,
                  kw_only: bool = MISSING):
    """
    Defines a dataclass field with a ``SkipIf`` condition.

    This function is a shortcut for ``dataclasses.field(...)``,
    adding metadata to specify a condition. If the condition
    evaluates to ``True``, the field is skipped during
    JSON serialization.

    Arguments:
        condition (Condition): The condition, if true skips serializing the field.
        default (Any): The default value for the field. Mutually exclusive with `default_factory`.
        default_factory (Callable[[], Any]): A callable to generate the default value.
                                             Mutually exclusive with `default`.
        init (bool): Include the field in the generated `__init__` method. Defaults to True.
        repr (bool): Include the field in the `__repr__` output. Defaults to True.
        hash (bool): Include the field in the `__hash__` method. Defaults to None.
        compare (bool): Include the field in comparison methods. Defaults to True.
        metadata (dict): Metadata to associate with the field. Defaults to None.
        kw_only (bool): If true, the field will become a keyword-only parameter to __init__().
    Returns:
        Field: A dataclass field with correct metadata set.

    Example:
        >>> from dataclasses import dataclass
        >>> @dataclass
        >>> class Example:
        >>>     my_str: str = skip_if_field(IS_NOT(True))
        >>> # Creates a condition which skips serializing `my_str`
        >>> # if its value `is not True`.
    """


class JSON:
    """
    Represents one or more mappings of JSON keys.

    See the docs on the :func:`json_key` function for more info.
    """
    __slots__ = ('keys',
                 'all',
                 'dump',
                 'path')

    keys: tuple[str, ...] | PathType
    all: bool
    dump: bool
    path: bool

    def __init__(self, *keys: str | PathPart, all=False, dump=True, path=False):
        ...


class JSONField(Field):
    """
    Alias to a :class:`dataclasses.Field`, but one which also represents a
    mapping of one or more JSON key names to a dataclass field.

    See the docs on the :func:`json_field` function for more info.
    """
    __slots__ = ('json', )

    json: JSON

    # In Python 3.10, dataclasses adds a new parameter to the :class:`Field`
    # constructor: `kw_only`
    #
    # Ref: https://docs.python.org/3.10/library/dataclasses.html#dataclasses.dataclass
    @overload
    def __init__(self, keys: _STR_COLLECTION, all: bool, dump: bool,
                 default, default_factory, init, repr, hash, compare,
                 metadata, path: bool = False):
        ...

    @overload
    def __init__(self, keys: _STR_COLLECTION, all: bool, dump: bool,
                 default, default_factory, init, repr, hash, compare,
                 metadata, path: bool = False):
        ...


# noinspection PyPep8Naming
def Pattern(pattern: str):
    """
    Represents a pattern (i.e. format string) for a date / time / datetime
    type or subtype. For example, a custom pattern like below::

        %d, %b, %Y %H:%M:%S.%f

    A sample usage of ``Pattern``, using a subclass of :class:`time`::

        time_field: Annotated[List[MyTime], Pattern('%I:%M %p')]

    :param pattern: A format string to be passed in to `datetime.strptime`
    """
    ...


class _PatternBase:
    """Base "subscriptable" pattern for date/time/datetime."""
    __slots__ = ()

    def __class_getitem__(cls, pattern: str) -> PatternedDT[date | time | datetime]:
        ...

    __getitem__ = _PatternBase.__class_getitem__


class DatePattern(date, _PatternBase):
    """
    An annotated type representing a date pattern (i.e. format string). Upon
    de-serialization, the resolved type will be a :class:`date` instead.

    See the docs on :func:`Pattern` for more info.
    """
    __slots__ = ()


class TimePattern(time, _PatternBase):
    """
    An annotated type representing a time pattern (i.e. format string). Upon
    de-serialization, the resolved type will be a :class:`time` instead.

    See the docs on :func:`Pattern` for more info.
    """
    __slots__ = ()


class DateTimePattern(datetime, _PatternBase):
    """
    An annotated type representing a datetime pattern (i.e. format string). Upon
    de-serialization, the resolved type will be a :class:`datetime` instead.

    See the docs on :func:`Pattern` for more info.
    """
    __slots__ = ()


class PatternedDT(Generic[DT]):
    """
    Base class for pattern matching using :meth:`datetime.strptime` when
    loading (de-serializing) a string to a date / time / datetime object.
    """

    # `cls` is the date/time/datetime type or subclass.
    # `pattern` is the format string to pass in to `datetime.strptime`.
    __slots__ = ('cls',
                 'pattern')

    cls: type[DT] | None
    pattern: str

    def __init__(self, pattern: str, cls: type[DT] | None = None):
        ...

    def get_transform_func(self) -> Callable[[str], DT]:
        """
        Build and return a load function which takes a `date_string` as an
        argument, and returns a new object of type :attr:`cls`.

        We try to parse the input string to a `cls` object in the following
        order:
            - In case it's an ISO-8601 format string, or a numeric timestamp,
              we first parse with the default load function (ex. as_datetime).
              We parse strings using the builtin :meth:`fromisoformat` method,
              as this is much faster than :meth:`datetime.strptime` - see link
              below for more details.
            - Next, we parse with :meth:`datetime.strptime` by passing in the
              :attr:`pattern` to match against. If the pattern is invalid, the
              method raises a ValueError, which is re-raised by our
              `Parser` implementation.

        Ref: https://stackoverflow.com/questions/13468126/a-faster-strptime

        :raises ValueError: If the input date string does not match the
          pre-defined pattern.
        """
        ...

    def __repr__(self):
        ...


class Container(list[T]):
    """Convenience wrapper around a collection of dataclass instances.

    For all intents and purposes, this should behave exactly as a `list`
    object.

    Usage:

        >>> from dataclass_wizard import Container, fromlist
        >>> from dataclasses import make_dataclass
        >>>
        >>> A = make_dataclass('A', [('f1', str), ('f2', int)])
        >>> list_of_a = fromlist(A, [{'f1': 'hello', 'f2': 1}, {'f1': 'world', 'f2': 2}])
        >>> c = Container[A](list_of_a)
        >>> print(c.prettify())

    """

    __slots__ = ('__dict__',
                 '__orig_class__')

    @cached_property
    def __model__(self) -> type[T]:
        """
        Given a declaration like Container[T], this returns the subscripted
        value of the generic type T.
        """
        ...

    def __str__(self):
        """
        Control the value displayed when ``print(self)`` is called.
        """
        ...

    def prettify(self, encoder: Encoder = json.dumps,
                 ensure_ascii=False,
                 **encoder_kwargs) -> str:
        """
        Convert the list of instances to a *prettified* JSON string.
        """
        ...

    def to_json(self, encoder: Encoder = json.dumps,
                **encoder_kwargs) -> str:
        """
        Convert the list of instances to a JSON string.
        """
        ...

    def to_json_file(self, file: str, mode: str = 'w',
                     encoder: FileEncoder = json.dump,
                     **encoder_kwargs) -> None:
        """
        Serializes the list of instances and writes it to a JSON file.
        """
        ...


class Condition:

    op: str         # Operator
    val: Any        # Value
    t_or_f: bool    # Truthy or falsy
    _wrapped: bool  # True if wrapped in `SkipIf()`

    def __init__(self, operator: str, value: Any):
        ...

    def __str__(self):
        ...

    def evaluate(self, other) -> bool:
        ...


# Aliases for conditions
# noinspection PyPep8Naming
def EQ(value: Any) -> Condition:
    """Create a condition for equality (==)."""


# noinspection PyPep8Naming
def NE(value: Any) -> Condition:
    """Create a condition for inequality (!=)."""


# noinspection PyPep8Naming
def LT(value: Any) -> Condition:
    """Create a condition for less than (<)."""


# noinspection PyPep8Naming
def LE(value: Any) -> Condition:
    """Create a condition for less than or equal to (<=)."""


# noinspection PyPep8Naming
def GT(value: Any) -> Condition:
    """Create a condition for greater than (>)."""


# noinspection PyPep8Naming
def GE(value: Any) -> Condition:
    """Create a condition for greater than or equal to (>=)."""


# noinspection PyPep8Naming
def IS(value: Any) -> Condition:
    """Create a condition for identity (is)."""


# noinspection PyPep8Naming
def IS_NOT(value: Any) -> Condition:
    """Create a condition for non-identity (is not)."""


# noinspection PyPep8Naming
def IS_TRUTHY() -> Condition:
    """Create a "truthy" condition for evaluation (if <var>)."""


# noinspection PyPep8Naming
def IS_FALSY() -> Condition:
    """Create a "falsy" condition for evaluation (if not <var>)."""


# noinspection PyPep8Naming
def SkipIf(condition: Condition) -> Condition:
    ...


SkipIfNone: Condition


def finalize_skip_if(skip_if: Condition,
                     operand_1: str,
                     conditional: str) -> str:
    ...


def get_skip_if_condition(skip_if: Condition,
                          _locals: dict[str, Any],
                          operand_2: str = None,
                          condition_i: int = None,
                          condition_var: str = '_skip_if_') -> 'str | bool':
    ...



import io
from dataclasses import dataclass
from typing import List, Optional, Dict

import pytest
from pytest_mock import MockerFixture

from dataclass_wizard.v0 import Container
from dataclass_wizard.v0.wizard_mixins import (
    JSONListWizard, JSONFileWizard, TOMLWizard, YAMLWizard
)
from .conftest import SampleClass


class MyListWizard(SampleClass, JSONListWizard):
    ...


class MyFileWizard(SampleClass, JSONFileWizard):
    ...


@dataclass
class MyYAMLWizard(YAMLWizard):
    my_str: str
    inner: Optional['Inner'] = None


@dataclass
class Inner:
    my_float: float
    my_list: List[str]


@pytest.fixture
def mock_open(mocker: MockerFixture):
    return mocker.patch('dataclass_wizard.v0.wizard_mixins.open')


def test_json_list_wizard_methods():
    """Test and coverage the base methods in JSONListWizard."""
    c1 = MyListWizard.from_json('{"f1": "hello", "f2": 111}')
    assert c1.__class__ is MyListWizard

    c2 = MyListWizard.from_json('[{"f1": "hello", "f2": 111}]')
    assert c2.__class__ is Container

    c3 = MyListWizard.from_list([{"f1": "hello", "f2": 111}])
    assert c3.__class__ is Container

    assert c2 == c3


def test_json_file_wizard_methods(mocker: MockerFixture, mock_open):
    """Test and coverage the base methods in JSONFileWizard."""
    filename = 'my_file.json'
    my_dict = {'f1': 'Hello world!', 'f2': 123}

    mock_decoder = mocker.Mock()
    mock_decoder.return_value = my_dict

    c = MyFileWizard.from_json_file(filename,
                                    decoder=mock_decoder)

    mock_open.assert_called_once_with(filename)
    mock_decoder.assert_called_once()

    mock_encoder = mocker.Mock()
    mock_open.reset_mock()

    c.to_json_file(filename,
                   encoder=mock_encoder)

    mock_open.assert_called_once_with(filename, 'w')
    mock_encoder.assert_called_once_with(my_dict, mocker.ANY)


def test_yaml_wizard_methods(mocker: MockerFixture):
    """Test and coverage the base methods in YAMLWizard."""
    yaml_data = """\
    my_str: test value
    inner:
        my_float: 1.2
        my_list:
            - hello, world!
            - 123\
    """

    # Patch open() to return a file-like object which returns our string data.
    m = mocker.patch('dataclass_wizard.v0.wizard_mixins.open',
                     mocker.mock_open(read_data=yaml_data))

    filename = 'my_file.yaml'

    obj = MyYAMLWizard.from_yaml_file(filename)

    m.assert_called_once_with(filename)
    m.reset_mock()

    assert obj == MyYAMLWizard(my_str='test value',
                               inner=Inner(my_float=1.2,
                                           my_list=['hello, world!', '123']))

    mock_open.return_value = mocker.mock_open()

    obj.to_yaml_file(filename)

    m.assert_called_once_with(filename, 'w')

    # default key casing for the dump process will be `lisp-case`
    m().write.assert_has_calls(
        [mocker.call('my-str'),
         mocker.call('inner'),
         mocker.call('my-float'),
         mocker.call('1.2'),
         mocker.call('my-list'),
         mocker.call('world!')],
        any_order=True)


def test_yaml_wizard_list_to_json():
    """Test and coverage the `list_to_json` method in YAMLWizard."""
    @dataclass
    class MyClass(YAMLWizard, key_transform='SNAKE'):
        my_str: str
        my_dict: Dict[int, str]

    yaml_string = MyClass.list_to_yaml([
        MyClass('42', {111: 'hello', 222: 'world'}),
        MyClass('testing!', {333: 'this is a test.'})
    ])

    assert yaml_string == """\
- my_dict:
    111: hello
    222: world
  my_str: '42'
- my_dict:
    333: this is a test.
  my_str: testing!
"""


def test_yaml_wizard_for_branch_coverage(mocker: MockerFixture):
    """
    For branching logic in YAMLWizard, mainly for code coverage purposes.
    """

    # This is to coverage the `if` condition in the `__init_subclass__`
    @dataclass
    class MyClass(YAMLWizard, key_transform=None):
        ...

    # from_yaml: To cover the case of passing in `decoder`
    mock_return_val = {'my_str': 'test string'}

    mock_decoder = mocker.Mock()
    mock_decoder.return_value = mock_return_val

    result = MyYAMLWizard.from_yaml('my stream', decoder=mock_decoder)

    assert result == MyYAMLWizard('test string')
    mock_decoder.assert_called_once()

    # to_yaml: To cover the case of passing in `encoder`
    mock_encoder = mocker.Mock()
    mock_encoder.return_value = mock_return_val

    m = MyYAMLWizard('test string')
    result = m.to_yaml(encoder=mock_encoder)

    assert result == mock_return_val
    mock_encoder.assert_called_once()

    # list_to_yaml: To cover the case of passing in `encoder`
    result = MyYAMLWizard.list_to_yaml([], encoder=mock_encoder)

    assert result == mock_return_val
    mock_encoder.assert_any_call([])


@dataclass
class MyTOMLWizard(TOMLWizard):
    my_str: str
    inner: Optional['Inner'] = None


def test_toml_wizard_methods(mocker: MockerFixture):
    """Test and cover the base methods in TOMLWizard."""
    toml_data = b"""\
my_str = "test value"
[inner]
my_float = 1.2
my_list = ["hello, world!", "123"]
    """

    # Mock open to return the TOML data as a string directly.
    mock_open = mocker.patch("dataclass_wizard.v0.wizard_mixins.open", mocker.mock_open(read_data=toml_data))

    filename = 'my_file.toml'

    # Test reading from TOML file
    obj = MyTOMLWizard.from_toml_file(filename)

    mock_open.assert_called_once_with(filename, 'rb')
    mock_open.reset_mock()

    assert obj == MyTOMLWizard(my_str="test value",
                               inner=Inner(my_float=1.2,
                                           my_list=["hello, world!", "123"]))

    # Test writing to TOML file
    # Mock open for writing to the TOML file.
    mock_open_write = mocker.mock_open()
    mocker.patch("dataclass_wizard.v0.wizard_mixins.open", mock_open_write)

    obj.to_toml_file(filename)

    mock_open_write.assert_called_once_with(filename, 'wb')


def test_toml_wizard_list_to_toml():
    """Test and cover the `list_to_toml` method in TOMLWizard."""
    @dataclass
    class MyClass(TOMLWizard, key_transform='SNAKE'):
        my_str: str
        my_dict: Dict[str, str]

    toml_string = MyClass.list_to_toml([
        MyClass('42', {'111': 'hello', '222': 'world'}),
        MyClass('testing!', {'333': 'this is a test.'})
    ])

    # print(toml_string)

    assert toml_string == """\
items = [
    { my_str = "42", my_dict = { 111 = "hello", 222 = "world" } },
    { my_str = "testing!", my_dict = { 333 = "this is a test." } },
]
"""


def test_toml_wizard_for_branch_coverage(mocker: MockerFixture):
    """Test branching logic in TOMLWizard, mainly for code coverage purposes."""

    # This is to cover the `if` condition in the `__init_subclass__`
    @dataclass
    class MyClass(TOMLWizard, key_transform=None):
        ...

    # from_toml: To cover the case of passing in `decoder`
    mock_return_val = {'my_str': 'test string'}

    mock_decoder = mocker.Mock()
    mock_decoder.return_value = mock_return_val

    result = MyTOMLWizard.from_toml('my stream', decoder=mock_decoder)

    assert result == MyTOMLWizard('test string')
    mock_decoder.assert_called_once()

    # to_toml: To cover the case of passing in `encoder`
    mock_encoder = mocker.Mock()
    mock_encoder.return_value = mock_return_val

    m = MyTOMLWizard('test string')
    result = m.to_toml(encoder=mock_encoder)

    assert result == mock_return_val
    mock_encoder.assert_called_once()

    # list_to_toml: To cover the case of passing in `encoder`
    result = MyTOMLWizard.list_to_toml([], encoder=mock_encoder)

    assert result == mock_return_val
    mock_encoder.assert_any_call({'items': []})



"""
Utility module for checking generic types provided by the `typing` library.
"""

__all__ = [
    'is_union',
    'get_origin',
    'get_origin_v2',
    'is_typed_dict_type_qualifier',
    'get_args',
    'get_keys_for_typed_dict',
    'is_typed_dict',
    'is_generic',
    'is_annotated',
    'eval_forward_ref',
    'eval_forward_ref_if_needed',
]

import functools
import sys
import typing

# noinspection PyUnresolvedReferences,PyProtectedMember
from typing import Union, _AnnotatedAlias  # type: ignore

from .._type_def import (
    FREF,
    PyForwardRef,
    PyNotRequired,
    PyReadOnly,
    PyRequired,
)
from ..constants import PY310_OR_ABOVE, PY313_OR_ABOVE
from ._string_conv import repl_or_with_union

# noinspection PyTypedDict
_TYPED_DICT_TYPE_QUALIFIERS = frozenset(
    {PyRequired, PyNotRequired, PyReadOnly}
)


def get_keys_for_typed_dict(cls):
    """
    Given a :class:`TypedDict` sub-class, returns a pair of
    (required_keys, optional_keys)
    """
    return cls.__required_keys__, cls.__optional_keys__


def _is_annotated(cls):
    return isinstance(cls, _AnnotatedAlias)


# Ref:
#   https://typing.readthedocs.io/en/latest/spec/typeddict.html#required-and-notrequired
#   https://typing.readthedocs.io/en/latest/spec/glossary.html#term-type-qualifier
def is_typed_dict_type_qualifier(cls) -> bool:
    return cls in _TYPED_DICT_TYPE_QUALIFIERS


# Ref:
#   https://github.com/python/typing/blob/master/typing_extensions/src_py3/typing_extensions.py#L2111
if PY310_OR_ABOVE:  # pragma: no cover
    from types import GenericAlias, UnionType
    _get_args = typing.get_args

    # noinspection PyUnresolvedReferences,PyProtectedMember
    _BASE_GENERIC_TYPES = (
        typing._GenericAlias,  # type: ignore
        typing._SpecialForm,
        GenericAlias,
        UnionType,
    )

    _UNION_TYPES = frozenset({
        UnionType,
        Union,
    })

    _TYPING_LOCALS = None

    def _process_forward_annotation(base_type):
        return PyForwardRef(base_type, is_argument=False)

    def is_union(cls) -> bool:
        return cls in _UNION_TYPES

    def get_origin_v2(cls):
        if type(cls) is UnionType:
            return UnionType

        return getattr(cls, '__origin__', cls)

    def _get_origin(cls, raise_=False):
        if isinstance(cls, UnionType):
            return Union

        try:
            return cls.__origin__
        except AttributeError:
            if raise_:
                raise
            return cls

else:  # pragma: no cover
    from typing_extensions import get_args as _get_args

    # noinspection PyProtectedMember,PyUnresolvedReferences
    _BASE_GENERIC_TYPES = (
        typing._GenericAlias,  # type: ignore
        typing._SpecialForm,
    )

    # PEP 585 is introduced in Python 3.9
    # PEP 604 (Allows writing union types as `X | Y`) is introduced
    #   in Python 3.10
    _TYPING_LOCALS = {'Union': Union}

    def _process_forward_annotation(base_type):
        return PyForwardRef(
            repl_or_with_union(base_type), is_argument=False)

    def is_union(cls) -> bool:
        return cls is Union

    def get_origin_v2(cls):
        return getattr(cls, '__origin__', cls)

    def _get_origin(cls, raise_=False):
        try:
            return cls.__origin__
        except AttributeError:
            if raise_:
                raise
            return cls


try:
    # noinspection PyProtectedMember,PyUnresolvedReferences
    from typing_extensions import _TYPEDDICT_TYPES, is_typeddict

except ImportError:
    from typing import is_typeddict as is_typed_dict

else:
    def is_typed_dict(cls: type) -> bool:
        """
        Checks if `cls` is a sub-class of ``TypedDict``
        """
        return isinstance(cls, _TYPEDDICT_TYPES)


def is_generic(cls):
    """
    Detects any kind of generic, for example `List` or `List[int]`. This
    includes "special" types like Union, Any ,and Tuple - anything that's
    subscriptable, basically.

    https://stackoverflow.com/a/52664522/10237506
    """
    return isinstance(cls, _BASE_GENERIC_TYPES)


get_args = _get_args
get_args.__doc__ = """
Get type arguments with all substitutions performed.

For unions, basic simplifications used by Union constructor are performed.
Examples::
    get_args(Dict[str, int]) == (str, int)
    get_args(int) == ()
    get_args(Union[int, Union[T, int], str][int]) == (int, str)
    get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
    get_args(Callable[[], T][int]) == ([], int)\
"""

# TODO refactor to use `typing.get_origin` when time permits.
get_origin = _get_origin
get_origin.__doc__ = """
Get the un-subscripted value of a type. If we're unable to retrieve this
value, return type `cls` if `raise_` is false.

This supports generic types, Callable, Tuple, Union, Literal, Final and
ClassVar. Return None for unsupported types.

Examples::

    get_origin(Literal[42]) is Literal
    get_origin(int) is int
    get_origin(ClassVar[int]) is ClassVar
    get_origin(Generic) is Generic
    get_origin(Generic[T]) is Generic
    get_origin(Union[T, int]) is Union
    get_origin(List[Tuple[T, T]][int]) == list

:raise AttributeError: When the `raise_` flag is enabled, and we are
  unable to retrieve the un-subscripted value.\
"""

is_annotated = _is_annotated
is_annotated.__doc__ = """Detects a :class:`typing.Annotated` class."""


if PY313_OR_ABOVE:
    # noinspection PyProtectedMember,PyUnresolvedReferences
    _eval_type = functools.partial(typing._eval_type, type_params=())
else:
    # noinspection PyProtectedMember,PyUnresolvedReferences
    _eval_type = typing._eval_type


def eval_forward_ref(base_type: FREF,
                     cls: type):
    """
    Evaluate a forward reference using the class globals, and return the
    underlying type reference.
    """

    if isinstance(base_type, str):
        base_type = _process_forward_annotation(base_type)

    # Evaluate the ForwardRef here
    base_globals = sys.modules[cls.__module__].__dict__

    return _eval_type(base_type, base_globals, _TYPING_LOCALS)


_ForwardRefTypes = frozenset(FREF.__constraints__)


def eval_forward_ref_if_needed(base_type: Union[type, FREF],
                               base_cls: type):
    """
    If needed, evaluate a forward reference using the class globals, and
    return the underlying type reference.
    """

    if type(base_type) in _ForwardRefTypes:
        # Evaluate the forward reference here.
        base_type = eval_forward_ref(base_type, base_cls)

    return base_type



__all__ = ['IdentityParser',
           'SingleArgParser',
           'Parser',
           'RecursionSafeParser',
           'PatternedDTParser',
           'LiteralParser',
           'UnionParser',
           'OptionalParser',
           'IterableParser',
           'TupleParser',
           'VariadicTupleParser',
           'NamedTupleParser',
           'NamedTupleUntypedParser',
           'MappingParser',
           'DefaultDictParser',
           'TypedDictParser']

from dataclasses import dataclass, InitVar, is_dataclass
from typing import (
    Type, Any, Optional, Tuple, Dict, Iterable, Callable, List
)

from .abstractions import AbstractParser
from .bases import AbstractMeta
from .class_helper import get_meta, _META
from .constants import TAG
from .errors import ParseError
from .models import PatternedDT, Extras
from .type_def import (
    FrozenKeys, NoneType, DefFactory,
    T, M, S, DD, LSQ, N, NT, DT
)
from .utils.typing_compat import (
    get_origin, get_args,
    get_keys_for_typed_dict, eval_forward_ref_if_needed)


# Type defs
GetParserType = Callable[[Type[T], Type, Extras], AbstractParser]
LoadHookType = Callable[[Any], T]
TupleOfParsers = Tuple[AbstractParser, ...]


@dataclass
class IdentityParser(AbstractParser[Type[T], T]):
    __slots__ = ()

    def __call__(self, o: Any) -> T:
        return o


@dataclass
class SingleArgParser(AbstractParser[Type[T], T]):
    __slots__ = ('hook', )

    hook: LoadHookType

    # noinspection PyDataclass
    def __post_init__(self, *_):
        if not self.hook:
            self.hook = lambda o: o

    def __call__(self, o: Any) -> T:
        return self.hook(o)


@dataclass
class Parser(AbstractParser[T, T]):
    __slots__ = ('hook', )

    hook: Callable[[Any, type[T]], T]

    def __call__(self, o: Any) -> T:
        return self.hook(o, self.base_type)


@dataclass
class RecursionSafeParser(AbstractParser):
    """
    Parser to handle cyclic or self-referential dataclasses.

    For example::

        @dataclass
        class A:
            a: A | None = None

        instance = fromdict(A, {'a': {'a': {'a': None}}})
    """
    __slots__ = ('extras', 'hook')

    extras: Extras
    hook: Optional[LoadHookType]

    def load_hook_func(self) -> LoadHookType:
        from .loaders import load_func_for_dataclass

        return load_func_for_dataclass(
            self.base_type,
            is_main_class=False,
            config=self.extras['config']
        )

    # TODO: decorating `load_hook_func` with `@cached_property` could
    #   be an alternate, bit cleaner approach.
    def __call__(self, o: Any) -> T:
        load_hook = self.hook

        if load_hook is None:
            load_hook = self.hook = self.load_hook_func()

        return load_hook(o)


@dataclass
class LiteralParser(AbstractParser[M, M]):
    __slots__ = ('value_to_type', )

    base_type: type[M]

    # noinspection PyDataclass
    def __post_init__(self, *_):
        self.value_to_type = {
            val: type(val) for val in get_args(self.base_type)
        }

    def __contains__(self, item) -> bool:
        """
        Return true if the LiteralParser is expected to handle the specified item
        type. Checks that the item is incorporated in the given expected values of
        the Literal.
        """
        return item in self.value_to_type

    def __call__(self, o: Any) -> M:
        """
        Checks for Literal equivalence, as mentioned here:
          https://www.python.org/dev/peps/pep-0586/#equivalence-of-two-literals

        """
        try:
            type_does_not_match = type(o) is not self.value_to_type[o]

        except KeyError:
            # No such Literal with the value of `o`
            e: Exception = ValueError('Value not in expected Literal values')
            raise ParseError(
                e, o, self.base_type, 'load',
                allowed_values=list(self.value_to_type))

        else:
            # The value of `o` is in the ones defined for the Literal, but
            # also confirm the type matches the one defined for the Literal.
            if type_does_not_match:
                expected_val = next(v for v in self.value_to_type if v == o)    # pragma: no branch
                e = TypeError(
                    'Value did not match expected type for the Literal')

                raise ParseError(
                    e, o, self.base_type, 'load',
                    have_type=type(o),
                    desired_type=self.value_to_type[o],
                    desired_value=expected_val,
                    allowed_values=list(self.value_to_type))

        return o


@dataclass
class PatternedDTParser(AbstractParser[PatternedDT, DT]):
    __slots__ = ('hook', )

    base_type: PatternedDT

    # noinspection PyDataclass
    def __post_init__(self, _cls: Type, extras: Extras, *_):
        if not isinstance(self.base_type, PatternedDT):
            dt_cls = self.base_type
            self.base_type = extras['pattern']
            self.base_type.cls = dt_cls

        self.hook = self.base_type.get_transform_func()

    def __call__(self, date_string: str) -> DT:
        try:
            return self.hook(date_string)
        except ValueError as e:
            raise ParseError(
                e, date_string, self.base_type.cls, 'load',
                pattern=self.base_type.pattern
            )


@dataclass
class OptionalParser(AbstractParser[T, Optional[T]]):
    __slots__ = ('parser', )

    get_parser: InitVar[GetParserType]

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):

        self.parser: AbstractParser = getattr(
            p := get_parser(self.base_type, cls, extras),
            '__call__', p
        )

    def __contains__(self, item):
        """Check if parser is expected to handle the specified item type."""
        if type(item) is NoneType:
            return True

        return super().__contains__(item)

    def __call__(self, o: Any) -> Optional[T]:
        if o is None:
            return o

        return self.parser(o)


@dataclass
class UnionParser(AbstractParser[Tuple[Type[T], ...], Optional[T]]):
    __slots__ = ('parsers', 'tag_to_parser', 'tag_key')

    base_type: Tuple[Type[T], ...]
    get_parser: InitVar[GetParserType]

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):

        # Tag key to search for when a dataclass is in a `Union` with
        # other types.
        config = extras.get('config')
        if config:
            self.tag_key: str = config.tag_key or TAG
            auto_assign_tags = config.auto_assign_tags
        else:
            self.tag_key = TAG
            auto_assign_tags = False

        parsers_list = []
        self.tag_to_parser = {}

        for t in self.base_type:
            t = eval_forward_ref_if_needed(t, cls)
            if t is not NoneType:
                parser = get_parser(t, cls, extras)

                if isinstance(parser, AbstractParser):
                    parsers_list.append(parser)

                elif is_dataclass(t):
                    meta = get_meta(t)
                    tag = meta.tag
                    if not tag and (auto_assign_tags or meta.auto_assign_tags):
                        cls_name = t.__name__
                        tag = cls_name
                        # We don't want to mutate the base Meta class here
                        if meta is AbstractMeta:
                            from .bases_meta import BaseJSONWizardMeta
                            cls_dict = {'__slots__': (), 'tag': tag}
                            # noinspection PyTypeChecker
                            meta: type[M] = type(cls_name + 'Meta', (BaseJSONWizardMeta, ), cls_dict)
                            _META[t] = meta
                        else:
                            meta.tag = cls_name
                    if tag:
                        # TODO see if we can use a mapping of dataclass type to
                        #   load func (maybe one passed in to __post_init__),
                        #   rather than generating one on the fly like this.
                        self.tag_to_parser[tag] = parser

        self.parsers = tuple(parsers_list)

    def __contains__(self, item):
        """Check if parser is expected to handle the specified item type."""
        return type(item) in self.base_type

    def __call__(self, o: Any) -> Optional[T]:
        if o is None:
            return o

        for parser in self.parsers:
            if o in parser:
                return parser(o)

        # Attempt to parse to the desired dataclass type, using the "tag"
        # field in the input dictionary object.
        try:
            tag = o[self.tag_key]
        except (TypeError, KeyError):
            # Invalid type (`o` is not a dictionary object) or no such key.
            pass
        else:
            try:
                return self.tag_to_parser[tag](o)
            except KeyError:
                raise ParseError(
                    TypeError('Object with tag was not in any of Union types'),
                    o, [p.base_type for p in self.parsers],
                    'load',
                    input_tag=tag,
                    tag_key=self.tag_key,
                    valid_tags=list(self.tag_to_parser.keys()))

        raise ParseError(
            TypeError('Object was not in any of Union types'),
            o, [p.base_type for p in self.parsers],
            'load',
            tag_key=self.tag_key
        )


@dataclass
class IterableParser(AbstractParser[Type[LSQ], LSQ]):
    """
    Parser for a :class:`list`, :class:`set`, :class:`frozenset`,
    :class:`deque`, or a subclass of either type.
    """
    __slots__ = ('hook',
                 'elem_parser')

    base_type: Type[LSQ]
    hook: Callable[[Iterable, Type[LSQ], AbstractParser], LSQ]
    get_parser: InitVar[GetParserType]

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):

        # Get the subscripted element type
        #   ex. `List[str]` -> `str`
        try:
            elem_type, = get_args(self.base_type)
        except ValueError:
            elem_type = Any

        # Base type of the object which is instantiable
        #   ex. `List[str]` -> `list`
        self.base_type = get_origin(self.base_type)

        self.elem_parser = getattr(
            p := get_parser(elem_type, cls, extras), '__call__', p,
        )

    def __call__(self, o: Iterable) -> LSQ:
        """
        Load an object `o` into a new object of type `base_type`.

        See the declaration of :var:`LSQ` for more info.
        """
        return self.hook(o, self.base_type, self.elem_parser)


@dataclass
class TupleParser(AbstractParser[Type[S], S]):
    """
    Parser for subscripted and un-subscripted :class:`Tuple`'s.

    See :class:`VariadicTupleParser` for the parser that handles the variadic
    form, i.e. ``Tuple[str, ...]``
    """
    __slots__ = ('hook',
                 'elem_parsers',
                 'total_count',
                 'required_count',
                 'elem_types')

    # Base type of the object which is instantiable
    #   ex. `Tuple[bool, int]` -> `tuple`
    base_type: Type[S]
    hook: Callable[[Any, Type[S], Optional[TupleOfParsers]], S]
    get_parser: InitVar[GetParserType]

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):

        # Get the subscripted values
        #   ex. `Tuple[bool, int]` -> (bool, int)
        self.elem_types = elem_types = get_args(self.base_type)
        self.base_type = get_origin(self.base_type)
        # A collection with a parser for each type argument
        elem_parsers = tuple(get_parser(t, cls, extras)
                             for t in elem_types)
        # Total count is generally the number of type arguments to `Tuple`, but
        # can be `Infinity` when a `Tuple` appears in its un-subscripted form.
        self.total_count: N = len(elem_parsers) or float('inf')
        # Minimum number of *required* type arguments
        #   Check for the count of parsers which don't handle `NoneType` -
        #   this should exclude the parsers for `Optional` or `Union` types
        #   that have `None` in the list of args.
        self.required_count: int = len(tuple(p for p in elem_parsers
                                             if not isinstance(p, AbstractParser)
                                             or None not in p))

        self.elem_parsers = elem_parsers or None

    def __call__(self, o: S) -> S:
        """
        Load an object `o` into a new object of type `base_type` (generally a
        :class:`tuple` or a sub-class of one)
        """
        # Confirm that the number of arguments in `o` matches the count in the
        # typed annotation.
        if not self.required_count <= len(o) <= self.total_count:
            e = TypeError('Wrong number of elements.')
            if self.required_count != self.total_count:
                desired_count = f'{self.required_count} - {self.total_count}'
            else:
                desired_count = str(self.total_count)

            # self.elem_parsers can be None at this moment
            elem_parsers_types = [getattr(p, 'base_type', tp) for p, tp in
                                  zip(self.elem_parsers, self.elem_types)] \
                if self.elem_parsers else self.elem_types

            raise ParseError(
                e, o, elem_parsers_types, 'load',
                desired_count=desired_count,
                actual_count=len(o))

        return self.hook(o, self.base_type, self.elem_parsers)


@dataclass
class VariadicTupleParser(TupleParser):
    """
    Parser that handles the variadic form of :class:`Tuple`'s,
    i.e. ``Tuple[str, ...]``

    Per `PEP 484`_, only **one** required type is allowed before the
    ``Ellipsis``. That is, ``Tuple[int, ...]`` is valid whereas
    ``Tuple[int, str, ...]`` would be invalid. `See here`_ for more info.

    .. _PEP 484: https://www.python.org/dev/peps/pep-0484/
    .. _See here: https://github.com/python/typing/issues/180

    """
    __slots__ = ('first_elem_parser', )

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):

        # Get the subscripted values
        #   ex. `Tuple[str, ...]` -> (str, )
        elem_types = get_args(self.base_type)
        # Base type of the object which is instantiable
        #   ex. `Tuple[bool, int]` -> `tuple`
        self.base_type = get_origin(self.base_type)
        # A one-element tuple containing the parser for the first type
        # argument.
        # Given `Tuple[T, ...]`, we only need a parser for `T`
        self.first_elem_parser: Tuple[AbstractParser]
        self.first_elem_parser = get_parser(elem_types[0], cls, extras),
        # Total count should be `Infinity` here, since the variadic form
        # accepts any number of possible arguments.
        self.total_count: N = float('inf')
        self.required_count = 0

    def __call__(self, o: M) -> M:
        """
        Load an object `o` into a new object of type `base_type` (generally a
        :class:`tuple` or a sub-class of one)
        """
        self.elem_parsers = self.first_elem_parser * len(o)
        return super().__call__(o)


@dataclass
class NamedTupleParser(AbstractParser[tuple, NT]):
    __slots__ = ('hook',
                 'field_to_parser',
                 'field_parsers')

    hook: Callable[
        [Any, type[tuple], Optional['FieldToParser'], List[AbstractParser]],
        NT
    ]
    get_parser: InitVar[GetParserType]

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):

        # Get the field annotations for the `NamedTuple` type
        type_anns: Dict[str, type[T]] = self.base_type.__annotations__

        self.field_to_parser: Optional['FieldToParser'] = {
            f: getattr(p := get_parser(ftype, cls, extras), '__call__', p)
            for f, ftype in type_anns.items()
        }

        self.field_parsers = list(self.field_to_parser.values())

    def __call__(self, o: Any) -> NT:
        """
        Load a dictionary or list to a `NamedTuple` sub-class (or an
        un-annotated `namedtuple`)
        """
        return self.hook(o, self.base_type,
                         self.field_to_parser, self.field_parsers)


@dataclass
class NamedTupleUntypedParser(AbstractParser[tuple, NT]):
    __slots__ = ('hook',
                 'dict_parser',
                 'list_parser')

    hook: Callable[[Any, Type[tuple], AbstractParser, AbstractParser], NT]
    get_parser: InitVar[GetParserType]

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):

        self.dict_parser = get_parser(dict, cls, extras).__call__
        self.list_parser = get_parser(list, cls, extras).__call__

    def __call__(self, o: Any) -> NT:
        """
        Load a dictionary or list to a `NamedTuple` sub-class (or an
        un-annotated `namedtuple`)
        """
        return self.hook(o, self.base_type,
                         self.dict_parser, self.list_parser)


@dataclass
class MappingParser(AbstractParser[Type[M], M]):
    __slots__ = ('hook',
                 'key_parser',
                 'val_parser',
                 'val_type')

    base_type: Type[M]
    hook: Callable[[Any, Type[M], AbstractParser, AbstractParser], M]
    get_parser: InitVar[GetParserType]

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):
        try:
            key_type, val_type = get_args(self.base_type)
        except ValueError:
            key_type = val_type = Any

        # Base type of the object which is instantiable
        #   ex. `Dict[str, Any]` -> `dict`
        self.base_type: Type[M] = get_origin(self.base_type)
        self.val_type = val_type

        val_parser = get_parser(val_type, cls, extras)

        self.key_parser = getattr(p := get_parser(key_type, cls, extras), '__call__', p)
        self.val_parser = getattr(val_parser, '__call__', val_parser)

    def __call__(self, o: M) -> M:
        return self.hook(o, self.base_type, self.key_parser, self.val_parser)


@dataclass
class DefaultDictParser(MappingParser[DD]):
    __slots__ = ('default_factory', )

    # Override the type annotations here
    base_type: Type[DD]
    hook: Callable[
        [Any, Type[DD], DefFactory, AbstractParser, AbstractParser], DD]

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):
        super().__post_init__(cls, extras, get_parser)

        # The default factory argument to pass to the `defaultdict` subclass
        val_type = self.val_type
        val_base_type = getattr(val_type, '__origin__', val_type)
        self.default_factory: DefFactory = val_base_type

    def __call__(self, o: DD) -> DD:
        return self.hook(o, self.base_type, self.default_factory,
                         self.key_parser, self.val_parser)


@dataclass
class TypedDictParser(AbstractParser[Type[M], M]):
    __slots__ = ('hook',
                 'key_to_parser',
                 'required_keys',
                 'optional_keys')

    base_type: Type[M]
    hook: Callable[[Any, Type[M], 'FieldToParser', FrozenKeys, FrozenKeys], M]
    get_parser: InitVar[GetParserType]

    def __post_init__(self, cls: Type,
                      extras: Extras,
                      get_parser: GetParserType):

        self.key_to_parser: 'FieldToParser' = {
            k: getattr(p := get_parser(v, cls, extras), '__call__', p)
            for k, v in self.base_type.__annotations__.items()
        }

        self.required_keys, self.optional_keys = get_keys_for_typed_dict(
            self.base_type
        )

    def __call__(self, o: M) -> M:
        try:
            return self.hook(o, self.base_type, self.key_to_parser,
                             self.required_keys, self.optional_keys)

        except KeyError as e:
            err: Exception = KeyError(f'Missing required key: {e.args[0]}')
            raise ParseError(err, o, self.base_type, 'load')

        except Exception:
            if not isinstance(o, dict):
                err = TypeError('Incorrect type for object')
                raise ParseError(
                    err, o, self.base_type, 'load', desired_type=self.base_type)
            else:
                raise



from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Callable, ClassVar, Literal

from ._decorators import cached_class_property
from .constants import TAG

if TYPE_CHECKING:  # pragma: no cover
    from datetime import tzinfo

    from ._bases import TypeToHook
    from ._bases_meta import PreDecoder
    from ._path_util import EnvFilePaths, SecretsDirs
    from ._type_def import META, FrozenKeys
    from .conditions import Condition
    from .enums import (
        DateTimeTo,
        EnvKeyStrategy,
        EnvPrecedence,
        KeyAction,
        KeyCase,
    )


class ABCOrAndMeta(type):
    """
    Metaclass to add class-level :meth:`__or__` and :meth:`__and__` methods
    to a base class of type :type:`M`.

    Ref:
      - https://stackoverflow.com/q/15008807/10237506
      - https://stackoverflow.com/a/57351066/10237506
    """

    def __or__(cls: META, other: META) -> META:
        """
        Merge two Meta configs. Priority will be given to the source config
        present in `cls`, e.g. the first operand in the '|' expression.

        Use case: Merge the Meta configs for two separate dataclasses into a
        single, unified Meta config.
        """
        src = cls
        src_dict = src.__dict__
        other_dict = other.__dict__

        base_dict = {'__slots__': ()}

        # Set meta attributes here.
        if src is AbstractMeta or src is AbstractEnvMeta:
            # Here we can't use `src` because the `bind_to` method isn't
            # defined on the abstract class. Use `other` instead, which
            # *will* be a concrete subclass of `AbstractMeta`.
            src = other
            # noinspection PyTypeChecker
            for k in src.fields_to_merge:
                if k in other_dict:
                    base_dict[k] = other_dict[k]
        else:
            # noinspection PyTypeChecker
            for k in src.fields_to_merge:
                if k in src_dict:
                    base_dict[k] = src_dict[k]
                elif k in other_dict:
                    base_dict[k] = other_dict[k]

        # This mapping won't be updated. Use the src by default.
        for k in src.__special_attrs__:
            if k in src_dict:
                base_dict[k] = src_dict[k]

        new_cls_name = src.__name__
        # Check if the type of the class we want to create is
        # `JSONWizard.Meta` or a subclass. If so, we want to avoid the
        # mandatory `__init_subclass__` call that gets invoked when creating
        # a new class, so use the superclass type instead.
        if src.__is_inner_meta__:
            # In a reversed MRO, the inheritance tree looks like this:
            #   |___ object -> BaseMeta
            #               -> AbstractMeta
            #               -> BaseJSONWizardMeta -> ...
            # So here, we want to choose the third-to-last class in the list.
            # noinspection PyUnresolvedReferences
            src = src.__mro__[-4]

        # noinspection PyTypeChecker
        return type(new_cls_name, (src, ), base_dict)

    def __and__(cls: META, other: META) -> META:
        """
        Merge the `other` Meta config into the first one, i.e. `cls`. This
        operation does not create a new class, but instead it modifies the
        source config `cls` in-place; the source will be the first operand in
        the '&' expression.

        Use case: Merge a separate Meta config (for a single dataclass) with
        the first config.
        """
        other_dict = other.__dict__

        # Set meta attributes here.
        # noinspection PyTypeChecker
        for k in cls.all_fields:
            if k in other_dict:
                setattr(cls, k, other_dict[k])

        return cls


class BaseMeta(metaclass=ABCOrAndMeta):
    """
    Base (shared) Meta definition.
    """
    __slots__ = ()

    # A list of class attributes that are exclusive to the Meta config.
    # When merging two Meta configs for a class, these are the only
    # attributes which will *not* be merged.
    __special_attrs__ = frozenset({
        'recursive',
        # 'debug',
        'field_to_alias',
        'field_to_alias_dump',
        'field_to_alias_load',
        'tag',
    })

    # Class attribute which enables us to detect a `EnvWizard.Meta` subclass.
    __is_inner_meta__ = False

    # When enabled, a specified Meta config for the main dataclass (i.e. the
    # class on which `from_dict` and `to_dict` is called) will cascade down
    # and be merged with the Meta config for each *nested* dataclass; note
    # that during a merge, priority is given to the Meta config specified on
    # each class.
    #
    # The default behavior is True, so the Meta config (if provided) will
    # apply in a recursive manner.
    recursive: ClassVar[bool] = True

    # The field name that identifies the tag for a class.
    #
    # When set to a value, an :attr:`TAG` field will be populated in the
    # dictionary object in the dump (serialization) process. When loading
    # (or de-serializing) a dictionary object, the :attr:`TAG` field will be
    # used to load the corresponding dataclass, assuming the dataclass field
    # is properly annotated as a Union type, ex.:
    #   my_data: Union[Data1, Data2, Data3]
    tag: ClassVar[str | None] = None

    # The dictionary key that identifies the tag field for a class. This is
    # only set when the `tag` field or the `auto_assign_tags` flag is enabled
    # in the `Meta` config for a dataclass.
    #
    # Defaults to '__tag__' if not specified.
    tag_key: ClassVar[str] = TAG

    # Auto-assign the class name as a dictionary "tag" key, for any dataclass
    # fields which are in a `Union` declaration, ex.:
    #   my_data: Union[Data1, Data2, Data3]
    auto_assign_tags: ClassVar[bool] = False

    # Determines whether we should we skip / omit fields with default values
    # (based on the `default` or `default_factory` argument specified for
    # the :func:`dataclasses.field`) in the serialization process.
    skip_defaults: ClassVar[bool] = False

    # Determines the :class:`Condition` to skip / omit dataclass
    # fields in the serialization process.
    skip_if: ClassVar[Condition | None] = None

    # Determines the condition to skip / omit fields with default values
    # (based on the `default` or `default_factory` argument specified for
    # the :func:`dataclasses.field`) in the serialization process.
    skip_defaults_if: ClassVar[Condition | None] = None

    # Enable Debug mode for more verbose log output.
    #
    # This setting can be a `bool`, `int`, or `str`:
    # - `True` enables debug mode with default verbosity.
    # - A `str` or `int` specifies the minimum log level (e.g., 'DEBUG', 10).
    #
    # Debug mode provides additional helpful log messages, including:
    # - Logging unknown JSON keys encountered during `from_dict` or `from_json`.
    # - Detailed error messages for invalid types during unmarshalling.
    #
    # Note: Enabling Debug mode may have a minor performance impact.
    debug: ClassVar[bool | int | str] = False

    # Custom load hooks for extending type support.
    #
    # Mapping: {Type -> hook}
    #
    # A hook must accept either:
    #   - one positional argument (runtime hook): value -> object
    #   - two positional arguments (codegen
    #     hook): (TypeInfo, Extras) -> str | TypeInfo
    #
    # The hook is invoked when loading a value annotated with the given type.
    type_to_load_hook: ClassVar[TypeToHook | None] = None

    # Custom dump hooks for extending type support.
    #
    # Mapping: {Type -> hook}
    #
    # A hook must accept either:
    #   - one positional argument (runtime hook): object -> JSON-serializable
    #     value
    #   - two positional arguments (codegen
    #     hook): (TypeInfo, Extras) -> str | TypeInfo
    #
    # The hook is invoked when dumping a value whose runtime type matches
    # the given type.
    type_to_dump_hook: ClassVar[TypeToHook | None] = None

    # ``pre_decoder``: Optional hook called before type loading.
    # Receives the container type plus (cls, TypeInfo, Extras) and may
    # return a transformed ``TypeInfo`` (e.g., wrapped in a function
    # which decodes JSON/delimited strings into list/dict for env
    # loading). Returning the input value leaves behavior unchanged.
    #
    #  Pre-decoder signature:
    #   (cls, container_tp, tp, extras) -> new_tp
    pre_decoder: ClassVar[PreDecoder] = None

    # Specifies the letter case used for JSON keys during serialization.
    #
    # This setting determines how dataclass field names are transformed
    # when generating keys in the output JSON object.
    #
    # By default, field names are emitted in `snake_case`.
    #
    # The setting is case-insensitive and supports shorthand assignment,
    # such as using the string 'P' instead of 'PASCAL'.
    #
    # If unset, this value defaults to `case` when provided.
    dump_case: ClassVar[KeyCase | str | None] = None

    # A custom mapping of dataclass fields to their JSON aliases (keys) used
    # during serialization only.
    #
    # Values may be a single alias string or a sequence of alias strings.
    # When a sequence is provided, the first alias is used as the output key.
    #
    # When set, this mapping overrides `field_to_alias` for dump behavior
    # only.
    field_to_alias_dump: ClassVar[
        Mapping[str, str | Sequence[str]] | None
    ] = None

    # Unsafe: Enables parsing of dataclasses in unions without requiring
    # the presence of a `tag_key`, i.e., a dictionary key identifying the
    # tag field in the input. Defaults to False.
    unsafe_parse_dataclass_in_union: ClassVar[bool] = False

    # Specifies how :class:`datetime` (and :class:`time`, where applicable)
    # objects are serialized during output.
    #
    # This setting controls how temporal values are emitted when converting
    # a dataclass to a Python dictionary (`to_dict`) or a JSON string
    # (`to_json`). It applies to serialization only and does not affect
    # deserialization.
    #
    # By default, values are serialized using ISO 8601 string format.
    #
    # Supported values are defined by :class:`DateTimeTo`.
    dump_date_time_as: ClassVar[DateTimeTo | str | None] = None

    # Specifies the timezone to assume for naive :class:`datetime` values
    # during serialization.
    #
    # By default, naive datetimes are rejected to avoid ambiguous or
    # environment-dependent behavior.
    #
    # When set, naive datetimes are interpreted as being in the specified
    # timezone before conversion to a UTC epoch timestamp.
    #
    # Common usage:
    #     assume_naive_datetime_tz = timezone.utc
    #
    # This setting applies to serialization only and does not affect
    # deserialization.
    assume_naive_datetime_tz: ClassVar[tzinfo | None] = None

    # Controls how `typing.NamedTuple` and `collections.namedtuple`
    # fields are loaded and serialized.
    #
    # - False (DEFAULT): load from list/tuple and serialize
    #                     as a positional list.
    # - True: load from mapping and serialize as a dict
    #           keyed by field name.
    #
    # In strict mode, inputs that do not match the selected mode
    # raise TypeError.
    #
    # Note:
    #   This option enforces strict shape matching for performance reasons.
    namedtuple_as_dict: ClassVar[bool | None] = None

    # If True (default: False), ``None`` is coerced to an empty
    # string (``""``)
    # when loading ``str`` fields.
    #
    # When False, ``None`` is coerced using ``str(value)``, so ``None``
    # becomes the literal string ``'None'`` for ``str`` fields.
    #
    # For ``Optional[str]`` fields, ``None`` is preserved by default.
    coerce_none_to_empty_str: ClassVar[bool | None] = None

    # Controls how leaf (non-recursive) types are detected during
    # serialization.
    #
    # - "exact" (DEFAULT): only exact built-in leaf types are treated
    #   as leaf values.
    # - "issubclass": subclasses of leaf types are also treated
    #   as leaf values.
    #
    # Leaf types are returned without recursive traversal. Bytes are still
    # handled separately according to their serialization rules.
    #
    # Note:
    #     The default "exact" mode avoids treating third-party scalar-like
    #     objects (e.g. NumPy scalars) as built-in leaf types.
    leaf_handling: ClassVar[Literal['exact', 'issubclass'] | None] = None

    # noinspection PyMethodParameters
    @cached_class_property
    def all_fields(cls: type) -> FrozenKeys:
        """Return a list of all class attributes"""
        keys = {}
        for base in reversed(cls.__mro__[:-1]):  # drop object, then reverse
            keys.update(base.__annotations__)
        return frozenset(keys)

    # noinspection PyMethodParameters
    @cached_class_property
    def fields_to_merge(cls) -> FrozenKeys:
        """Return a list of class attributes, minus `__special_attrs__`"""
        return cls.all_fields - cls.__special_attrs__


class AbstractMeta(BaseMeta):
    """
    Base class definition for the `JSONWizard.Meta` inner class.
    """
    __slots__ = ()

    # A list of class attributes that are exclusive to the Meta config.
    # When merging two Meta configs for a class, these are the only
    # attributes which will *not* be merged.
    __special_attrs__ = frozenset({
        'recursive',
        # 'debug',
        'field_to_alias',
        'field_to_alias_dump',
        'field_to_alias_load',
        'tag',
    })

    # Class attribute which enables us to detect a `JSONWizard.Meta` subclass.
    __is_inner_meta__ = False

    # Specifies the letter case to use for JSON keys when both loading and
    # dumping.
    #
    # This is a convenience setting that applies the same key casing rule to
    # both deserialization (load) and serialization (dump).
    #
    # If set, it is used as the default for both `load_case` and
    # `dump_case`, unless either is explicitly specified.
    #
    # The setting is case-insensitive and supports shorthand assignment,
    # such as using the string 'C' instead of 'CAMEL'.
    case: ClassVar[KeyCase | str | None] = None

    # Specifies the letter case used to match JSON keys when mapping them
    # to dataclass fields during deserialization.
    #
    # This setting determines how dataclass field names are transformed
    # when looking up corresponding keys in the input JSON object. It does
    # not affect keys in `TypedDict` or `NamedTuple` subclasses.
    #
    # By default, JSON keys are assumed to be in `snake_case`, and fields
    # are matched directly without transformation.
    #
    # The setting is case-insensitive and supports shorthand assignment,
    # such as using the string 'C' instead of 'CAMEL'.
    #
    # If set to `A` or `AUTO`, all supported key casing transforms are
    # attempted at runtime, and the resolved transform is cached for
    # subsequent lookups.
    #
    # If unset, this value defaults to `case` when provided.
    load_case: ClassVar[KeyCase | str | None] = None

    # A custom mapping of dataclass fields to their JSON aliases (keys).
    #
    # Values may be a single alias string or a sequence of alias strings.
    #
    # - During deserialization (load), any listed alias for a field is
    #   accepted.
    # - During serialization (dump), the first alias is used by default.
    #
    # This mapping overrides default key casing and implicit field-to-key
    # transformations (e.g., "my_field" → "myField") for the affected fields.
    #
    # This setting applies to both load and dump unless explicitly overridden
    # by `field_to_alias_load` or `field_to_alias_dump`.
    field_to_alias: ClassVar[Mapping[str, str | Sequence[str]] | None] = None

    # A custom mapping of dataclass fields to their JSON aliases (keys) used
    # during deserialization only.
    #
    # Values may be a single alias string or a sequence of alias strings.
    # Any listed alias is accepted when mapping input JSON keys to
    # dataclass fields.
    #
    # When set, this mapping overrides `field_to_alias` for load behavior
    # only.
    field_to_alias_load: ClassVar[
        Mapping[str, str | Sequence[str]] | None] = None

    # Defines the action to take when an unknown JSON key is encountered
    # during `from_dict` or `from_json` calls. An unknown key is one that
    # does not map to any dataclass field.
    #
    # Valid options are:
    # - `"ignore"` (default): Silently ignore unknown keys.
    # - `"warn"`: Log a warning for each unknown key. Requires `debug`
    #   to be `True` and properly configured logging.
    # - `"raise"`: Raise an `UnknownKeyError` for the first unknown key
    #   encountered.
    on_unknown_key: ClassVar[KeyAction | None] = None

    @classmethod
    def bind_to(cls, dataclass: type, create=True, is_default=True):
        """
        Initialize hook which applies the Meta config to `dataclass`, which is
        typically a subclass of :class:`JSONWizard`.

        :param dataclass: A class which has been decorated by the `@dataclass`
          decorator; typically this is a sub-class of :class:`JSONWizard`.
        :param create: When true, a separate loader/dumper will be created
          for the class. If disabled, this will access the root loader/dumper,
          so modifying this should affect global settings across all
          dataclasses that use the JSON load/dump process.
        :param is_default: When enabled, the Meta will be cached as the
          default Meta config for the dataclass. Defaults to true.

        """
        raise NotImplementedError


class AbstractEnvMeta(BaseMeta):
    """
    Base class definition for the `EnvWizard.Meta` inner class.
    """
    __slots__ = ()

    # A list of class attributes that are exclusive to the Meta config.
    # When merging two Meta configs for a class, these are the only
    # attributes which will *not* be merged.
    __special_attrs__ = frozenset({
        'recursive',
        'debug',
        'field_to_env_load',
        'field_to_alias_dump',
        'tag',
    })

    # Class attribute which enables us to detect a `EnvWizard.Meta` subclass.
    __is_inner_meta__ = False

    # `True` to load environment variables from an `.env` file, or a
    # list/tuple of dotenv files.
    #
    # This can also be set to a path to a custom dotenv file, for example:
    #   `path/to/.env.prod`
    #
    # Simply passing in a filename such as `.env.prod` will search the current
    # directory, as well as any parent folders (working backwards to the root
    # directory), until it locates the given file.
    #
    # If multiple files are passed in, later files in the list/tuple will take
    # priority over earlier files.
    #
    # For example, in below the '.env.last' file takes priority over '.env':
    #   env_file = '.env', '.env.last'
    env_file: ClassVar[EnvFilePaths] = None

    # Prefix for all environment variables. Defaults to `None`.
    env_prefix: ClassVar[str | None] = None

    # secrets_dir: The secret files directory or a sequence of directories.
    # Defaults to `None`.
    secrets_dir: ClassVar[SecretsDirs] = None

    # The key lookup strategy to use for Env Var Names.
    #
    # The default strategy is `SCREAMING_SNAKE_CASE` > `snake_case`.
    load_case: ClassVar[EnvKeyStrategy | str | None] = None

    # Environment Precedence (order) to search for values
    # Defaults to EnvPrecedence.SECRETS_ENV_DOTENV
    env_precedence: ClassVar[EnvPrecedence | None] = None

    # A custom mapping of dataclass fields to their env vars (keys) used
    # during deserialization only.
    #
    # Values may be a single alias string or a sequence of alias strings.
    # Any listed alias is accepted when mapping input env vars to
    # dataclass fields.
    field_to_env_load: ClassVar[
        Mapping[str, str | Sequence[str]] | None] = None

    # Defines the action to take when an unknown JSON key is encountered
    # during `from_dict` or `from_json` calls. An unknown key is one
    # that does not map to any dataclass field.
    #
    # Valid options are:
    # - `"ignore"` (default): Silently ignore unknown keys.
    # - `"warn"`: Log a warning for each unknown key. Requires `debug`
    #   to be `True` and properly configured logging.
    # - `"raise"`: Raise an `UnknownKeyError` for the first unknown key
    #   encountered.
    # on_unknown_key: ClassVar[KeyAction] = None

    @classmethod
    def bind_to(cls, env_class: type, create=True, is_default=True):
        """
        Initialize hook which applies the Meta config to `env_class`,
        which is typically a subclass of :class:`EnvWizard`.

        :param env_class: A sub-class of :class:`EnvWizard`.
        :param create: When true, a separate loader/dumper will be created
          for the class. If disabled, this will access the root loader/dumper,
          so modifying this should affect global settings across all
          dataclasses that use the JSON load/dump process.
        :param is_default: When enabled, the Meta will be cached as the
          default Meta config for the dataclass. Defaults to true.

        """
        raise NotImplementedError


class _BaseHookRegistry:
    __slots__ = ()
    __HOOKS__: ClassVar[dict[type, Callable]]

    @classmethod
    def register_hook(cls, typ: type, func: Callable):
        cls.__HOOKS__[typ] = func

    @classmethod
    def get_hook(cls, typ: type) -> Callable | None:
        return cls.__HOOKS__.get(typ)


class BaseLoadHook(_BaseHookRegistry):
    """
    Container class for load type hooks.
    """


class BaseDumpHook(_BaseHookRegistry):
    """
    Container class for dump type hooks.
    """
